// -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
//
// Time-stamp: <2000-09-11 15:24:28 haoboy>
//
// Copyright (c) 2000 by the University of Southern California
// All rights reserved.
//
// Permission to use, copy, modify, and distribute this software and its
// documentation in source and binary forms for non-commercial purposes
// and without fee is hereby granted, provided that the above copyright
// notice appear in all copies and that both the copyright notice and
// this permission notice appear in supporting documentation. and that
// any documentation, advertising materials, and other materials related
// to such distribution and use acknowledge that the software was
// developed by the University of Southern California, Information
// Sciences Institute.  The name of the University may not be used to
// endorse or promote products derived from this software without
// specific prior written permission.
//
// THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
// the suitability of this software for any purpose.  THIS SOFTWARE IS
// PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Other copyrights might apply to parts of this software and are so
// noted when applicable.
//
// Original source contributed by Gaeil Ahn. See below.
//
// $Header: /home/ip-qos/REPOSITORY/ns/mns_v2.0/classifier-addr-mpls.cc,v 1.3 2001/11/21 13:30:15 ip-qos Exp $

// XXX
//
// - Because MPLS header contains pointers, it cannot be used WITH multicast 
//   routing which replicates packets
// - Currently it works only with flat routing.

/**************************************************************************
 * Copyright (c) 2000 by Gaeil Ahn                                   	  *
 * Everyone is permitted to copy and distribute this software.		  *
 * Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute    *
 * this sources.							  *
 **************************************************************************/

/************************************************************
 *                                                          *
 *    File: File for packet switching in MPLS node          *
 *    Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Dec. 1999      *
 *                                                          *
 ************************************************************/

#include "common/packet.h"
#include "trace/trace.h"
#include "classifier-addr-mpls.h"

int hdr_mpls::offset_;

static class shimhdreaderClass : public PacketHeaderClass {
public:
	shimhdreaderClass() : PacketHeaderClass("PacketHeader/MPLS", 
						sizeof(hdr_mpls)) {
		bind_offset(&hdr_mpls::offset_);
	}
} class_shimhdreader;

static class MPLSAddrClassifierClass : public TclClass {
public:
	MPLSAddrClassifierClass() : TclClass("Classifier/Addr/MPLS") {}
	virtual TclObject* create(int, const char*const*) {
		return (new MPLSAddressClassifier());
	}
	virtual void bind();
	virtual int method(int argc, const char*const* argv);
} class_mpls_addr_classifier;

void MPLSAddrClassifierClass::bind()
{
	TclClass::bind();
	add_method("minimum-lspid");
	add_method("dont-care");
	add_method("ordered-control?");
	add_method("on-demand?");
	//add_method("enable-ordered-control");
	//add_method("enable-on-demand");
}

int MPLSAddrClassifierClass::method(int ac, const char*const* av)
{
	Tcl& tcl = Tcl::instance();
	int argc = ac - 2;
	const char*const* argv = av + 2;

	if (argc == 2) {
		if (strcmp(argv[1], "minimum-lspid") == 0) {
			tcl.resultf("%d", MPLS_MINIMUM_LSPID);
			return (TCL_OK);
		} else if (strcmp(argv[1], "dont-care") == 0) {
			tcl.resultf("%d", MPLS_DONTCARE);
			return (TCL_OK);
		}
	}
	return TclClass::method(ac, av);
}

MPLSAddressClassifier::MPLSAddressClassifier() : 
	data_driven_(0), control_driven_(0), on_demand_(0), ordered_control_(0)
{
	dont_care_  = MPLS_DONTCARE;
	PFT_.NB_    = 0;
	ERB_.NB_    = 0;
	LIB_.NB_    = 0;
	ttl_        = 32;
	trace_mpls_ = 0;
}

void MPLSAddressClassifier::delay_bind_init_all()
{
	delay_bind_init_one("ttl_");
	delay_bind_init_one("exp_");
	delay_bind_init_one("trace_mpls_");   
	delay_bind_init_one("label_");
	delay_bind_init_one("enable_reroute_");
	delay_bind_init_one("failure_detection_");
	delay_bind_init_one("reroute_option_");
	delay_bind_init_one("data_driven_");
	delay_bind_init_one("control_driven_");
	delay_bind_init_one("dont_care_");
	delay_bind_init_one("is_cbq_installed_");
	delay_bind_init_one("reroute_egress_lsr_");
	AddressClassifier::delay_bind_init_all();
}

// Arguments: varName, localName, tracer
int MPLSAddressClassifier::delay_bind_dispatch(const char *vn, 
					       const char* ln, TclObject *t)
{
	if (delay_bind(vn, ln, "ttl_", &ttl_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "exp_", &exp_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "trace_mpls_", &trace_mpls_, t)) 
		return TCL_OK;
	if (delay_bind(vn, ln, "label_", &label_, t)) 
		return TCL_OK;
	if (delay_bind(vn, ln, "enable_reroute_", &enable_reroute_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "failure_detection__", &failure_detection_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "reroute_option_", &reroute_option_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "data_driven_", &data_driven_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "control_driven_", &control_driven_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "dont_care_", &dont_care_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "is_cbq_installed_", &is_cbq_installed_, t))
		return TCL_OK;
	if (delay_bind(vn, ln, "reroute_egress_lsr_", &reroute_egress_lsr_, t))
		return TCL_OK;
	return AddressClassifier::delay_bind_dispatch(vn, ln, t);
}

int MPLSAddressClassifier::classify(Packet* p)
{
	int LIBnb = -1;
	
	int nexthop = MPLSclassify(p, LIBnb);
	
	if ( enable_reroute_== 1 && LIBnb >=0 ) {
		if ( LIB_.Entry_[LIBnb].error_ == 1 ||
		     failure_detection_ == 1 ) {
			nexthop = do_reroute(p, LIBnb);
		}
		else if ( is_link_down(nexthop) ) {
			PI_.shimhdr_ = GetShimHeader(p);
			int iLabel = PI_.shimhdr_->label_;
			PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
			trace("L",size_, iLabel,"Drop(linkFail)", -1, -1, -1);
			return(-10);
		}
	}
	
	PacketScheduler(p, LIBnb);
	
	if (nexthop == MPLS_GOTO_L3)
		return AddressClassifier::classify(p);
	
	// XXX Do NOT return -1, which lets the classifier to 
	// process this packet twice!!
	return (nexthop == -1) ? Classifier::ONCE : nexthop;
}

void MPLSAddressClassifier::PacketScheduler(Packet *const p, int LIBnb)
{
	int tmp, QoSid=-1;
	shim_header* shimhdr;
	
	hdr_ip* iphdr = hdr_ip::access(p);
	
	shimhdr = GetShimHeader(p);
	
	if (is_cbq_installed_ == 0) return;
	
	if (LIBnb >= 0 && is_cbq_installed_ == 1) {
		InvsERBlookup(LIBnb, tmp, tmp, QoSid);
		if (QoSid >= 0) {
			iphdr->flowid() = QoSid;
		}
		else {
			iphdr->flowid() = MPLS_SBTflowID + ExpValue_;
		}
	}
	else {
		iphdr->flowid() = shimhdr->exp_;
	}
}

int MPLSAddressClassifier::is_link_down(int node)
{
	Tcl& tcl = Tcl::instance();   
	tcl.evalf("[%s set mpls_mod_] get-link-status %d", name(), node);
	return (strcmp(tcl.result(), "down") == 0) ? 1 : 0;
}

int MPLSAddressClassifier::do_reroute(Packet* p, int &LIBnb)
{
	int oIface, oLabel, LIBptr;
	PI_.shimhdr_ = GetShimHeader(p);
	int iLabel = PI_.shimhdr_->label_;
	
	if ( !SecLSPLookup(PI_.dst_.addr_, PI_.phb_, LIBnb, oIface, oLabel, LIBptr) ) {
		if ( size_ > 0 )
			return ( convertL2toL2(iLabel, oIface, oLabel, LIBptr, LIBnb) );
		else
			return ( convertL3toL2(oIface, oLabel, LIBptr, LIBnb) );
	}
	if ( !aPathLookup(PI_.dst_.addr_, PI_.phb_ ,LIBnb, oIface, oLabel, LIBptr) ) {
		if ( size_ > 0 )
			return ( convertL2toL2(iLabel, oIface, oLabel, LIBptr, LIBnb) );
		else
			return ( convertL3toL2(oIface, oLabel, LIBptr, LIBnb) );
	}
	else {
		LIBnb = -1;
		PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		
		switch (reroute_option_) {
		case MPLS_L3FORWARDING:
			trace("L",size_, iLabel,"L3(linkFail)", -1, -1, -1);
			return MPLS_GOTO_L3;
		default:
			trace("L",size_, iLabel,"Drop(linkFail)", -1, -1, -1);
			return -10;
		}
	}
}

int MPLSAddressClassifier::convertL3toL2(int oIface, int oLabel, int LIBptr, int &LIBnb)
{
	int iLabel= -1;
	
	while (oLabel >= 0) {
		/* penultimate hop */
		if (oLabel == 0)  {
			/* no operation */
			trace("U",size_, iLabel, "Push(penultimate)", 
			      oIface, oLabel, ttl_);
		} else {
			/* push operation in ingerss LSR */
			PI_.shimhdr_ = push(PI_.shimhdr_, LIBnb);
			trace("U",size_, iLabel, "Push(ingress)", 
			      oIface, oLabel, ttl_);
		}
		if (LIBptr >= 0) {
			/* stack operation */
			iLabel = oLabel;
			LIBnb = LIBptr;
			LIBlookup(LIBnb, oIface, oLabel, LIBptr);
		} else
			break;
	}

	if (oLabel < 0) {
		LIBnb = -1;
		if (size_ > 0)
			PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		trace("U",size_, iLabel , "L3(errorLabel)", -1,-1, -1);
		return MPLS_GOTO_L3;
	}
	if (oIface < 0) {
		LIBnb = -1;
		PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		trace("U", size_, iLabel, "L3(errorOIF)", -1 , -1, -1);
		return MPLS_GOTO_L3;
	} else
		// Guaranteed returned oIface is >= 0 or MPLS_GOTO_L3
		return oIface;
}

int MPLSAddressClassifier::convertL2toL2(int iLabel, int oIface, 
					 int oLabel, int LIBptr, int &LIBnb)
{
	int  ttl = PI_.shimhdr_->ttl_;
	
	// push(stack) operation after swap or pop
	if (oLabel == 0) {
		// in penultimate hop
		PI_.shimhdr_ = pop(PI_.shimhdr_);
		trace("L", size_, iLabel, "Pop(penultimate)", 
		      oIface, oLabel, ttl);
	} else if (oLabel > 0) {
		// swap operation 
		swap(PI_.shimhdr_,oLabel);
		trace("L", size_, iLabel, "Swap", oIface, oLabel, ttl);
	} else {
		// Errored Label
		LIBnb = -1;
		PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		trace("L", size_, iLabel, "L3(errorLabel)", -1, -1, -1);
		return MPLS_GOTO_L3;
	}
	while (LIBptr >= 0) {
		// stack operation
		iLabel= oLabel;
		LIBnb = LIBptr;
		LIBlookup(LIBnb, oIface, oLabel, LIBptr);
		PI_.shimhdr_ = push(PI_.shimhdr_, LIBnb);
		trace("L", size_, iLabel, "Push(tunnel)", 
		      oIface, oLabel, ttl_);
	}   
	if (oIface < 0) {  
		LIBnb = -1;
		if (size_ > 0)
			PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		trace("L",size_, iLabel , "L3(errorOIf)", -1 , -1, -1);
		return MPLS_GOTO_L3;
	}
	// Guaranteed returned oIface >= 0
	return oIface;
}

// Process unlabeled packet
int MPLSAddressClassifier::processIP(int &LIBnb)
{
	int iLabel, oIface, oLabel, LIBptr;
	iLabel = LIBnb = -1;

	// Insert code to manipulate PHB
	if ( !PFTlookup(PI_.dst_.addr_, PI_.phb_, oIface, oLabel, LIBptr, LIBnb) )
		// Push operation
		return convertL3toL2(oIface, oLabel, LIBptr, LIBnb);
	
	// L3 forwarding
	// Traffic-driven, triggered by MPLS switch
	if (data_driven_) 
		Tcl::instance().evalf("%s ldp-trigger-by-switch %d", 
				      name(), PI_.dst_.addr_);
	trace("U", size_, iLabel, "L3", -1, -1, -1);               
	return MPLS_GOTO_L3;
}

// Process labeled packet
int MPLSAddressClassifier::processLabelP(int &LIBnb)
{
	int oIface,oLabel,LIBptr;
	int iLabel = PI_.shimhdr_->label_;
	
	PI_.shimhdr_ = checkTTL(PI_.shimhdr_);
	
	if (size_ == 0)
		// TTL check
		return MPLS_GOTO_L3;
	
	if ( !LIBlookup(-1, iLabel, oIface, oLabel, LIBptr, LIBnb) ) {
		// splicing check
		while ( oIface==-1 && oIface==-1 && LIBptr >=0) {
			int sLIBptr = LIBptr;
			LIBnb = LIBptr;
			LIBlookup(sLIBptr, oIface, oLabel, LIBptr);
		}
		// Label swapping operation 
		return convertL2toL2(iLabel, oIface,
				     oLabel, LIBptr, LIBnb);
	}
	else {
		PI_.shimhdr_ = DelAllShimHeader(PI_.shimhdr_);
		trace("L",size_, iLabel,"L3(errorLabel)", -1, -1, -1);
		return ( MPLS_GOTO_L3 );
	}
}

int MPLSAddressClassifier::MPLSclassify(Packet* p, int &LIBnb)
{
	GetIPInfo(p, PI_.dst_, PI_.phb_, PI_.srcnode_);
	PI_.shimhdr_ = GetShimHeader(p);
	
	// XXX Using header size to determine if this is a MPLS-labeled packet
	// is a very bad method. We should have some explicit flag that labels
	// every packet; this flag will only be set on for every MPLS-labeled
	// packet. This can be done by a bitmap field in the common header.
	if (size_ == 0) 
		// Unlabeled packet
		return processIP(LIBnb);

	// Labeled packet 
	int ret = processLabelP(LIBnb);
	if (ret == MPLS_GOTO_L3) {  
		PI_.shimhdr_ = GetShimHeader(p);
		return processIP(LIBnb);
	}
	return ret;   
}

shim_header* MPLSAddressClassifier::checkTTL(shim_header* shimhdr)
{
	shimhdr->ttl_--;
	int ttl   = shimhdr->ttl_;
	int iLabel= shimhdr->label_;
   
	if (ttl == 0) {
		shimhdr = DelAllShimHeader(shimhdr);
		trace("L", size_, iLabel, "L3(TTL=0)", -1, -1, ttl);
	}
	return shimhdr;
}

void MPLSAddressClassifier::GetIPInfo(Packet* p, ns_addr_t &dst,
				      int &phb, int &srcnode)
{
	hdr_ip* iphdr = hdr_ip::access(p);
	dst = iphdr->dst_;
	srcnode = iphdr->src_.addr_;
	phb = iphdr->flowid();
}

shim_header* MPLSAddressClassifier::GetShimHeader(Packet* p)
{
	hdr_mpls* mplshdr = hdr_mpls::access(p);
	shim_header* sptr;
	sptr = &(mplshdr->sh[0]);
	size_ = 0;
	if ((mplshdr->sh[0].label_ == 0) &&
	    (mplshdr->sh[0].bflag_ == 0) && 
	    (mplshdr->sh[0].ttl_ == 0)) {
		mplshdr->sh[0].bflag_ = -1;
		mplshdr->sh[0].label_ = -1;
		mplshdr->sh[0].ttl_   = -1;
		
		hdr_ip* iphdr = hdr_ip::access(p);
		mplshdr->sh[0].exp_ = iphdr->flowid();
		
		for (int i=1; i < MPLS_MaxShimHdrNB; i++)
			mplshdr->sh[i].bflag_ = -1;
	} else {
		for (int i=1; mplshdr->sh[i].bflag_ != -1; i++) {
			sptr = &(mplshdr->sh[i]);
			size_ += 4;
		}
	}
	return (sptr);
}

shim_header* MPLSAddressClassifier::DelAllShimHeader(shim_header* shimhdr)
{
	while (shimhdr->bflag_ != -1)
		shimhdr = pop(shimhdr);
	size_ = 0;
	return(shimhdr);
}

shim_header* MPLSAddressClassifier::push(shim_header* shimhdr, int LIBnb)
{
	shimhdr++;
	shimhdr->label_ = LIB_.Entry_[LIBnb].oLabel_;
	shimhdr->bflag_ = 0;
	shimhdr->ttl_ = ttl_;
	
	if (LIBnb < 0)
		shimhdr->exp_ = 0;
	else {
		int PFTnb = InvsPFTlocate(LIBnb);
		if ( PFTnb >= 0 )  shimhdr->exp_ = PFT_.Entry_[LIBnb].Prio_;
		else               shimhdr->exp_ = 0;
        }
	
	ExpValue_ = shimhdr->exp_; // the priority value of MPLS packet
	size_ += 4;
	
	return(shimhdr);
}

shim_header* MPLSAddressClassifier::pop(shim_header* shimhdr)
{
	ExpValue_ = shimhdr->exp_; // the priority value of MPLS packet
	shimhdr->bflag_ = -1;
	size_ -= 4;
	shimhdr--;
	
	return(shimhdr);
}

void MPLSAddressClassifier::swap(shim_header* shimhdr, int oLabel)
{
	ExpValue_ = shimhdr->exp_; // the priority value of MPLS packet
	shimhdr->label_ = oLabel;
}

void MPLSAddressClassifier::install(int slot, NsObject *target)
{
	Tcl& tcl = Tcl::instance();
	if ((slot >= 0) && (slot < nslot_) && 
	    (slot_[slot] != NULL)) {
		if (strcmp(slot_[slot]->name(), target->name()) != 0)
			tcl.evalf("%s routing-update %d %.15g",
				  name(), slot,
				  Scheduler::instance().clock());
		else
			tcl.evalf("%s routing-nochange %d %.15g",
				  name(), slot,
				  Scheduler::instance().clock());
	} else
		tcl.evalf("%s routing-new %d %.15g",
			  name(), slot,
			  Scheduler::instance().clock());
	Classifier::install(slot,target);
}

int MPLSAddressClassifier::command(int argc, const char*const* argv)
{
	char out[400];
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "control-driven?") == 0) {
			tcl.resultf("%d", control_driven_);
			return (TCL_OK);
		} else if (strcmp(argv[1], "data-driven?") == 0) {
			tcl.resultf("%d", data_driven_);
			return (TCL_OK);
		} else if (strcmp(argv[1], "ordered-control?") == 0) {
			tcl.resultf("%d", ordered_control_);
			return (TCL_OK);
		} else if (strcmp(argv[1], "on-demand?") == 0) {
			tcl.resultf("%d", on_demand_);
			return (TCL_OK);
		} else if (strcmp(argv[1], "enable-control-driven") == 0) {
			// XXX data-driven and control-driven triggers are
			// exclusive 
			control_driven_ = 1;
			data_driven_ = 0;
			return (TCL_OK);
		} else if (strcmp(argv[1], "enable-data-driven") == 0) {
			control_driven_ = 0;
			data_driven_ = 1;
			return (TCL_OK);
		} else if (strcmp(argv[1], "enable-ordered-control") == 0) {
			ordered_control_ = 1;
			return (TCL_OK);
		} else if (strcmp(argv[1], "enable-on-demand") == 0) {
			on_demand_ = 1;
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "PFTdump") == 0) {
			// <classifier> PFTdump <nodeid>
			PFTdump(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1], "ERBdump") == 0) {
			// <classifier> ERBdump <nodeid>
			ERBdump(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1], "LIBdump") == 0) {
			// <classifier> LIBdump <nodeid>
			LIBdump(argv[2]);
			return (TCL_OK);
		}
		
		int LSPid  = atoi(argv[2]);
		int LIBptr = -1;
		int ERBnb  = ERBlocate(LSPid, -1, LIBptr);
		
		if (strcmp(argv[1], "get-fec-for-lspid") == 0) {
			// <classifier get-fec-for-lspid <LSPid>
			if (ERBnb >= 0)
				tcl.resultf("%d", ERB_.Entry_[ERBnb].FEC_);
			else   
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-qosid-for-lspid") == 0) {
			// <classifier get-qosid-for-lspid <LSPid>
			if (ERBnb >=0)
				tcl.resultf("%d", ERB_.Entry_[ERBnb].QoSid_);
			else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-lspid-for-qosid") == 0) {
			// <classifier get-lspid-for-qosid <qosid>
			int QoSid = atoi(argv[2]);
			ERBnb = ERBlocate(QoSid);
			if (ERBnb >=0)
				tcl.resultf("%d", ERB_.Entry_[ERBnb].LSPid_);
			else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-tunnelid-for-lspid") == 0) {
			// <classifier get-tunnelid-for-lspid <lspid>
			int tmp;
			int tunnelid = -1;
			if (ERBnb >=0) {
				LIBlookup(LIBptr, tmp, tmp, LIBptr);
				InvsERBlookup(LIBptr, tunnelid, tmp, tmp);
			}
			tcl.resultf("%d", tunnelid);
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-spliced-lsp") == 0) {
			// <classifier get-spliced-lsp <LSPid>
			char SplicedLSPidList[500];
			if ( ERBgetSplicedLSPid(atoi(argv[2]), SplicedLSPidList) > 0 )
				tcl.result(SplicedLSPidList);
			else // there is no spliced LSP
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-failnext-for-lspid") == 0) {
			// <classifier get-failnext-for-lspid <lspid>
			int tmp;
			ERBnb = ERBlocate(LSPid, -1, tmp);
			if ( (ERBnb > -1) &&
			     (ERB_.Entry_[ERBnb].FailNext_[0] != '*') )
				sprintf(out, "%s", ERB_.Entry_[ERBnb].FailNext_);
			else
				sprintf(out, "%d",-1);
			tcl.result(out);                  
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-faildst-for-lspid") == 0) {
			// <classifier get-faildst-for-lspid <lspid>
			int tmp;
			int ERBnb = ERBlocate(LSPid, -1, tmp);
			if ( (ERBnb > -1) &&
			     (ERB_.Entry_[ERBnb].iIface_ > -1) )
				tcl.resultf("%d", ERB_.Entry_[ERBnb].iIface_);
			else
				tcl.result("-1");
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "exist-fec") == 0) {
			// <classifier> exist-fec <fec> <phb>
			int LIBptr;
			int PFTnb = PFTlocate(atoi(argv[2]),
					      atoi(argv[3]),
					      LIBptr);         
			if (PFTnb > -1)
				tcl.result("1");
			else
				tcl.result("-1");
			return (TCL_OK);
		} else if(strcmp(argv[1], "splice-lsp") == 0) {
			// <classifier> splice-lsp <Flspid> <Llspid>
			ERBsplice(atoi(argv[3]), atoi(argv[2]));
			return (TCL_OK);
		} else if(strcmp(argv[1], "erb-append-qosid") == 0) {
			// <classifier> erb-append-qosid <lspid> <qosid>
			int LSPid = atoi(argv[2]);
			int QoSid = atoi(argv[3]);
			ERBappendQoSid(LSPid, QoSid);
			return (TCL_OK);
		} else if(strcmp(argv[1], "erb-append-failnext") == 0) {
			// <classifier> erb-append-failnext <lspid> <???>
			int tmp;
			int ERBnb = ERBlocate(atoi(argv[2]), -1, tmp);
			if (ERBnb > -1) {
				delete [] ERB_.Entry_[ERBnb].FailNext_;
				ERB_.Entry_[ERBnb].FailNext_ = new char[strlen(argv[3])+1];
				strcpy(ERB_.Entry_[ERBnb].FailNext_,argv[3]);
			}
			return (TCL_OK);
		} else if (strcmp(argv[1], "aPathBinding") == 0) {
			// <classifier> aPathBinding <LSPid> <ReLSPid>
			int LSPid   = atoi(argv[2]);
			int ReLSPid = atoi(argv[3]);
			if ( !aPathBinding(LSPid, ReLSPid) )
				tcl.result("1");
			else 
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "sec-lsp-binding") == 0) {
			// <classifier> sec-lsp-binding <LSPid> <ReLSPid>
			int LSPid   = atoi(argv[2]);
			int ReLSPid = atoi(argv[3]);
			if ( !SecLSPBinding(LSPid, ReLSPid) )
				tcl.result("1");
			else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "get-lspid-for-fecphb") == 0) {
			// <classifier> get-lspid-for-fecphb <fec> <phb>
			int LIBptr;
			int LSPid;
			int tmp;
			PFTlocate(atoi(argv[2]), atoi(argv[3]), LIBptr);
			InvsERBlookup(LIBptr, LSPid, tmp, tmp);
			tcl.resultf("%d", LSPid);
			return (TCL_OK);
		} else if (strcmp(argv[1], "install") == 0) {
			// <classifier> install <slot> <target>
			int slot = atoi(argv[2]);
			NsObject* target = (NsObject*)TclObject::lookup(argv[3]);
			install(slot,target);
			return (TCL_OK);
		}
	} else if (argc == 5) {
		if (strcmp(argv[1], "ErLspBinding") == 0) {
			// <classifier> ErLspBinding <FEC> <PHB> <lspid>
			if ( !ErLspBinding(atoi(argv[2]), atoi(argv[3]),
					   -1, atoi(argv[4])) )
				tcl.result("1");
			else     
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "erb-append-reroute") == 0) {
			// <classifier> erb-append-reroute <LSPid> <iLabel> <iIface>
			ERBappendReroute(atoi(argv[2]), atoi(argv[3]), atoi(argv[4]));
			return (TCL_OK);
		}
		int LIBptr = -1;
		int iLabel, oLabel, iIface, oIface;
		if ( atoi(argv[4]) < 0)
			PFTlocate(atoi(argv[2]),atoi(argv[3]), LIBptr);
		else
			ERBlocate(atoi(argv[4]), -1, LIBptr);	
		
		if (strcmp(argv[1], "GetInIface") == 0) {
			// <classifier> GetInIface <FEC> <PHB> <LSPid>
			if (LIBptr > -1) {
				LIBgetIncoming(LIBptr,iIface,iLabel);
				tcl.resultf("%d", iIface);
			} else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "GetInLabel") == 0) {
			// <classifier> GetInLabel <FEC> <PHB> <LSPid>
			if (LIBptr > -1) {
				LIBgetIncoming(LIBptr, iIface, iLabel);
				tcl.resultf("%d", iLabel);
			} else 
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "GetOutIface") == 0) {
			// <classifier> GetOutIface <FEC> <PHB> <LSPid>
			if (LIBptr > -1) {
				LIBlookup(LIBptr, oIface, oLabel, LIBptr);
				tcl.resultf("%d", oIface);
			} else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "GetOutLabel") == 0) {
			// <classifier> GetOutLabel <FEC> <PHB> <LSPid>
			if (LIBptr > -1) {
				LIBlookup(LIBptr, oIface, oLabel, LIBptr);
				tcl.resultf("%d", oLabel);
			} else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "pft-append-prio") == 0) {
			// <classifier> pft-append-prio <FEC> <PHB> <Priority>
			int tmp;
			int PFTnb = PFTlocate(atoi(argv[2]),
					      atoi(argv[3]), tmp);
			
			if ( PFTnb > -1 ) {
				PFT_.Entry_[PFTnb].Prio_ = atoi(argv[4]);
				tcl.result("1");
			}
			else
				tcl.result("-1");
			return (TCL_OK);
		}
	} else if (argc == 6) {
		if (strcmp(argv[1], "ErLspStacking") == 0) {
			// <classifier> ErLspStacking <fec0> <erlspid0> <fec> <erfecid>
			int erfec0   = atoi(argv[2]);
			int erlspid0 = atoi(argv[3]);
			int erfec    = atoi(argv[4]);
			int erlspid  = atoi(argv[5]);
			if ( !ErLspStacking(erfec0, erlspid0, erfec, erlspid) )
				tcl.result("1");
			else     
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "FlowAggregation") == 0) {
			// <classifier> FlowAggregation <fineFEC> <finePHB>
			//              <coarseFEC> <coarsePHB>
			int fineaddr   = atoi(argv[2]);
			int finePHB    = atoi(argv[3]);
			int coarseaddr = atoi(argv[4]);
			int coarsePHB  = atoi(argv[5]);
			if ( !FlowAggregation(fineaddr, finePHB, coarseaddr, coarsePHB) )
				tcl.result("1");
			else     
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "aPathBinding") == 0) {
			// <classifier> aPathBinding <FEC> <PHB> 
			// <erFEC> <ReLSPid>
			int FEC     = atoi(argv[2]);
			int PHB     = atoi(argv[3]);
			int erFEC   = atoi(argv[4]);
			int ReLSPid = atoi(argv[5]);
			if ( !aPathBinding(FEC, PHB, erFEC, ReLSPid) )
				tcl.result("1");
			else     
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "sec-lsp-binding") == 0) {
			// <classifier> sec-lsp-binding <FEC> <PHB> <erFEC> <ReLSPid>
			int FEC     = atoi(argv[2]);
			int PHB     = atoi(argv[3]);
			int erFEC   = atoi(argv[4]);
			int ReLSPid = atoi(argv[5]);
			if ( !SecLSPBinding(FEC, PHB, erFEC, ReLSPid) )
				tcl.result("1");
			else
				tcl.result("-1");
			return (TCL_OK);
		} else if (strcmp(argv[1], "set-lib-error") == 0) {
			// <classifier> set-lib-error <fec> <phb> <lspid> <state>
			int LIBptr;
			if ( atoi(argv[4]) < 0 )
				PFTlocate(atoi(argv[2]), atoi(argv[3]), LIBptr);
			else
				ERBlocate(atoi(argv[4]), -1, LIBptr);
			if (LIBptr > -1)
				LIB_.Entry_[LIBptr].error_ = atoi(argv[5]);
			return (TCL_OK);
		}
	} else if (argc == 8) {      
		int addr  = atoi(argv[2]);
		int LSPid = atoi(argv[3]);
		int PHB   = LSPid;
		int LIBptr, PFTnb, ERBnb;
		if (LSPid == MPLS_DEFAULT_PHB)   // topology-based LSP
			PFTnb = PFTlocate(addr, PHB, LIBptr);
		else                             // ER-LSP
			ERBnb = ERBlocate(LSPid, addr, LIBptr);
		
		if (strcmp(argv[1], "LSPrelease") == 0) {
			// <classifier> LSPrelease <fec> <lspid> <iif> 
			// 	<ilabel> <oif> <olabel>
			if (LSPid < 0) {
				// Topology-based LSP
				if (PFTnb >= 0) {
					// PFT entry exist
					LIBupdate(LIBptr, atoi(argv[4]), 
						  atoi(argv[5]), atoi(argv[6]),
						  atoi(argv[7]));
					if ( !LIBisdeleted(LIBptr) )
						PFTdeleteLIBptr(LIBptr);
				}
			} else {
				// ER-LSP
				if ( ERBnb >= 0 ) {
					// ERB entries exist
					LIBupdate(LIBptr, atoi(argv[4]),
						  atoi(argv[5]), atoi(argv[6]),
						  atoi(argv[7]));
					if ( !LIBisdeleted(LIBptr) ) {
						int tmp;
						for (;;) {
							ERBnb = InvsERBlookup(LIBptr, tmp, tmp, tmp);
							if (ERBnb > -1) ERBdelete(ERBnb);
							else            break;
						}
						PFTdeleteLIBptr(LIBptr);
					}
				}
			}
			return (TCL_OK);        
		} else if (strcmp(argv[1], "LSPsetup") == 0) {
			// <classifier> LSPsetup <fec> <lspid> <iif> 
			//     <ilabel> <oif> <olabel>
			if (LSPid == MPLS_DEFAULT_PHB) {
				// Topology-based LSP
				if (PFTnb < 0) {
					// PFT entry does not exist
					int ptr = LIBinsert(atoi(argv[4]),
							    atoi(argv[5]),
							    atoi(argv[6]),
							    atoi(argv[7]));
					if (ptr > -1) {
						PFTinsert(addr, 
							  MPLS_DEFAULT_PHB,
							  ptr);
					} else {
						return (TCL_ERROR);
					}
				} else {
					// PFTnb >= 0
					// PFT entry already exists
					if (LIBptr > -1) {
						// check whether alternative path is setup
						if ( (enable_reroute_ == 1) &&
						     (LIB_.Entry_[LIBptr].oIface_ > -1) &&
						     (reroute_option_ == MPLS_MAKENEWLSP) &&
						     (atoi(argv[6]) > -1) && 
						     (is_link_down(LIB_.Entry_[LIBptr].oIface_)) ) {
							PFT_.Entry_[PFTnb].aPATHptr_ = 
								LIBinsert(atoi(argv[4]),atoi(argv[5]),
									  atoi(argv[6]),atoi(argv[7]));
							
							for (int i=0; i<LIB_.NB_; i++) {
								if ( LIB_.Entry_[i].oIface_ == atoi(argv[2]) ) {
									for (int k=0; k<PFT_.NB_; k++) {
										if (PFT_.Entry_[k].LIBptr_ == i)
											PFT_.Entry_[k].aPATHptr_ =
												PFT_.Entry_[PFTnb].aPATHptr_;
									}
								}
							}  
						}		
						else
							LIBupdate(LIBptr,atoi(argv[4]),atoi(argv[5]),
								  atoi(argv[6]),atoi(argv[7]));
					}
					else {
						int ptr = LIBinsert(atoi(argv[4]),atoi(argv[5]),
								    atoi(argv[6]),atoi(argv[7]));
						
						if (ptr > -1) PFTupdate(PFTnb, ptr);
						else          return (TCL_ERROR);	
					}
				}
			}
			else {
				// ER-LSP
				if (ERBnb < 0) {
					// ERB entry does not exist
					int ptr = LIBinsert(atoi(argv[4]),
							    atoi(argv[5]),
							    atoi(argv[6]),
							    atoi(argv[7]));
					if (ptr > -1) {
						ERBinsert(LSPid,addr,ptr);
					} else {
						return (TCL_ERROR);
					}
				} 
				else {
					// ERBnb >= 0
					// ERB entry already exists
					if (LIBptr > -1)
						LIBupdate(LIBptr, atoi(argv[4]), atoi(argv[5]),
							  atoi(argv[6]), atoi(argv[7]));
					else {
						int ptr = 
							LIBinsert(atoi(argv[4]),atoi(argv[5]),
								  atoi(argv[6]),atoi(argv[7]));
						if (ptr > -1)
							ERBupdate(ERBnb, ptr);
						else
							return (TCL_ERROR);
					}
				}
			}
			return (TCL_OK);
		}
	}
	
	return (AddressClassifier::command(argc, argv));
}

//--------------------------------------------------
// PFT (Partial Forwarding Table)
//--------------------------------------------------
int MPLSAddressClassifier::PFTnewEntry()
{
	for (int i=0; i<PFT_.NB_; i++) {
		if ( PFT_.Entry_[i].FEC_ < 0) {
			PFTdelete(i);
			return (i);
		}
	}
	if (PFT_.NB_ == MPLS_MaxPFTEntryNB-1) {
		printf("Error : excess MPLS_MaxPFTEntryNB\n");
		exit(0);
	}
	else {
		PFT_.NB_++;
		PFTdelete(PFT_.NB_-1);
		return(PFT_.NB_-1);
	}
}

int MPLSAddressClassifier::PFTinsert(int FEC, int PHB, int LIBptr)
{
	int newnb = PFTnewEntry();
	
	PFT_.Entry_[newnb].FEC_    = FEC;
	PFT_.Entry_[newnb].PHB_    = PHB;
	PFT_.Entry_[newnb].LIBptr_ = LIBptr;
	
	return(newnb);
}

void MPLSAddressClassifier::PFTdelete(int entrynb)
{
	PFT_.Entry_[entrynb].FEC_      = -1;
	PFT_.Entry_[entrynb].PHB_      = -1;
	PFT_.Entry_[entrynb].LIBptr_   = -1;
	PFT_.Entry_[entrynb].aPATHptr_ = -1;
	PFT_.Entry_[entrynb].SLIBptr_  = -1;
	PFT_.Entry_[entrynb].Prio_     =  0;
}

void MPLSAddressClassifier::PFTdeleteLIBptr(int LIBptr)
{
	for (int i = 0; i < PFT_.NB_; i++) {
		if (PFT_.Entry_[i].LIBptr_ == LIBptr)
			PFTdelete(i);
		if ( PFT_.Entry_[i].SLIBptr_ == LIBptr )
			PFT_.Entry_[i].SLIBptr_ = -1;
	}  
}

void MPLSAddressClassifier::PFTupdate(int entrynb, int LIBptr)
{
	PFT_.Entry_[entrynb].LIBptr_ = LIBptr;
}

int MPLSAddressClassifier::PFTlocate(int FEC, int PHB, int &LIBptr)
{
	LIBptr = -1;
	if (FEC < 0) return -1;
	
	for (int i = 0; i < PFT_.NB_; i++)
		if ((PFT_.Entry_[i].FEC_ == FEC) && (PFT_.Entry_[i].PHB_ == PHB)) {
			LIBptr = PFT_.Entry_[i].LIBptr_;
			return i;
		}
	return -1;
}

int MPLSAddressClassifier::PFTlookup(int FEC, int PHB,
				     int &oIface, int &oLabel,
				     int &LIBptr, int &LIBnb)
{
	oIface = oLabel = LIBptr = -1;
	if (FEC < 0) return -1;
	
	for (int i = 0; i < PFT_.NB_; i++)
		if ((PFT_.Entry_[i].FEC_ == FEC) &&
		    (PFT_.Entry_[i].PHB_ == PHB)) {
			LIBnb = PFT_.Entry_[i].LIBptr_;
			return(LIBlookup(LIBnb, oIface, oLabel, LIBptr));
		}
	
	for (int i = 0; i < PFT_.NB_; i++)
		if ( (PFT_.Entry_[i].FEC_ == FEC) &&
		     (PFT_.Entry_[i].PHB_ == MPLS_DEFAULT_PHB) ) {
			LIBnb = PFT_.Entry_[i].LIBptr_;
			return(LIBlookup(LIBnb, oIface, oLabel, LIBptr));
		}
	
	return -1;
}

int MPLSAddressClassifier::InvsPFTlocate(int LIBptr)
{
	for (int i=0; i<PFT_.NB_; i++)
		if ( PFT_.Entry_[i].LIBptr_ == LIBptr )
			return(i);
	
	return (-1);
}

void MPLSAddressClassifier::PFTdump(const char *id)
{
	fflush(stdout);
	printf("      --) ___PFT dump___ [node: %s] (--\n", id);
	printf("---------------------------------------------\n");
	printf("     FEC       Fid       PRio      LIBptr     SLIBptr   AlternativePath\n");
	for (int i = 0; i < PFT_.NB_; i++) {
		if (PFT_.Entry_[i].FEC_ == -1) continue;
		
		printf("     %d    ", PFT_.Entry_[i].FEC_);
		printf("     %d    ", PFT_.Entry_[i].PHB_);
		printf("     %d    ", PFT_.Entry_[i].Prio_);
		printf("     %d    ", PFT_.Entry_[i].LIBptr_);
		printf("     %d    ", PFT_.Entry_[i].SLIBptr_);
		printf("     %d\n"  , PFT_.Entry_[i].aPATHptr_);
	}
	printf("\n");
}

//--------------------------------------------------
// ER-LSP Table
//--------------------------------------------------
int MPLSAddressClassifier::ERBnewEntry()
{
	for (int i=0; i<ERB_.NB_; i++) {
		if ( ERB_.Entry_[i].LSPid_ < 0 && ERB_.Entry_[i].FEC_ < 0)
			return (i);
	}
	
	if (ERB_.NB_ == MPLS_MaxERBEntryNB-1) {
		printf("Error : excess MPLS_MaxERBEntryNB\n");
		exit(0);
	}
	else
		return(ERB_.NB_++);
}

int MPLSAddressClassifier::ERBinsert(int LSPid, int FEC, int LIBptr)
{
	int newnb = ERBnewEntry();
	
	ERB_.Entry_[newnb].LSPid_    = LSPid;
	ERB_.Entry_[newnb].FEC_      = FEC;
	ERB_.Entry_[newnb].LIBptr_   = LIBptr;
	ERB_.Entry_[newnb].FailNext_ = new char[2];
	strcpy(ERB_.Entry_[newnb].FailNext_,"*");
	ERB_.Entry_[newnb].QoSid_    =  -1;
	ERB_.Entry_[newnb].aPATHptr_ =  -1;
	ERB_.Entry_[newnb].iLabel_   =  -1;
	ERB_.Entry_[newnb].iIface_   =  -1;
	ERB_.Entry_[newnb].SLIBptr_  =  -1;
	
	return (newnb);	
}
 
void MPLSAddressClassifier::ERBappendQoSid(int LSPid, int QoSid)
{
    int LIBptr;
    int entrynb = ERBlocate(LSPid, -1, LIBptr);
    
    ERB_.Entry_[entrynb].QoSid_ = QoSid;
}

void MPLSAddressClassifier::ERBappendReroute(int LSPid, int iLabel, int iIface)
{
    int LIBptr;
    int entrynb = ERBlocate(LSPid, -1, LIBptr);
    ERB_.Entry_[entrynb].iLabel_ = iLabel;
    ERB_.Entry_[entrynb].iIface_ = iIface;
}

void MPLSAddressClassifier::ERBdelete(int entrynb)
{
	ERB_.Entry_[entrynb].FEC_      = -1;
	ERB_.Entry_[entrynb].LSPid_    = -1;
	ERB_.Entry_[entrynb].LIBptr_   = -1;
	ERB_.Entry_[entrynb].QoSid_    = -1;
	ERB_.Entry_[entrynb].aPATHptr_ = -1;
	ERB_.Entry_[entrynb].iLabel_   = -1;
	ERB_.Entry_[entrynb].iIface_   = -1;
	ERB_.Entry_[entrynb].SLIBptr_  = -1;
	delete [] ERB_.Entry_[entrynb].FailNext_;
}

void MPLSAddressClassifier::ERBdeleteRef(int entrynb)
{
	for (int i=0; i<PFT_.NB_; i++)
		if ( PFT_.Entry_[i].aPATHptr_ == entrynb )
			PFT_.Entry_[i].aPATHptr_ = -1;
	
	for (int i=0; i<ERB_.NB_; i++) {
		if ( ERB_.Entry_[i].aPATHptr_ == entrynb )
			ERB_.Entry_[i].aPATHptr_ = -1;
		
		if ( ERB_.Entry_[i].SLIBptr_ == entrynb )
			ERB_.Entry_[i].SLIBptr_ = -1;
        }
}

void MPLSAddressClassifier::ERBsplice(int fromLSPid, int toLSPid)
{
	int LIBptr;
	
	int ERBnb = ERBlocate(fromLSPid, -1, LIBptr);
	int newLIBnb = LIBnewEntry();
	
	LIB_.Entry_[newLIBnb].iIface_   = LIB_.Entry_[LIBptr].iIface_;
	LIB_.Entry_[newLIBnb].iLabel_   = LIB_.Entry_[LIBptr].iLabel_;
	LIB_.Entry_[newLIBnb].oIface_   = -1;
	LIB_.Entry_[newLIBnb].oLabel_   = -1;
	LIB_.Entry_[newLIBnb].LIBptr_   = LIBptr;
	
	int newERBnb = ERBnewEntry();    
	
	ERB_.Entry_[newERBnb].FEC_      = ERB_.Entry_[ERBnb].FEC_;
	ERB_.Entry_[newERBnb].LSPid_    = toLSPid;
	ERB_.Entry_[newERBnb].LIBptr_   = newLIBnb;
	
	ERB_.Entry_[newERBnb].QoSid_    =  ERB_.Entry_[ERBnb].QoSid_;
	
	ERB_.Entry_[newERBnb].FailNext_ = new char[2];
	strcpy(ERB_.Entry_[newERBnb].FailNext_,"*");
	ERB_.Entry_[newERBnb].aPATHptr_ =  -1;
	ERB_.Entry_[newERBnb].iLabel_   =  -1;
	ERB_.Entry_[newERBnb].iIface_   =  -1;
	ERB_.Entry_[newERBnb].SLIBptr_  =  -1;
}

int MPLSAddressClassifier::ERBgetSplicedLSPid(int LSPid, char *SplicedLSPid)
{
	
	if (LSPid < 0) return(-1);
	
	SplicedLSPid[0]='\0';
	int sLSPid, tmp;
	int found = -1;
	
	int LIBptr;
	int ERBnb = ERBlocate(LSPid, -1, LIBptr);
	if (LIBptr < 0) return(-1);
	
	for (int i=0; i<LIB_.NB_; i++)
		if ( (LIB_.Entry_[i].oLabel_ == -1) &&
		     (LIB_.Entry_[i].oIface_ == -1) &&
		     (LIB_.Entry_[i].LIBptr_ > -1) && 
		     (LIB_.Entry_[i].LIBptr_ == ERBnb) ) {
			
			if ( i == ERBnb ) continue;
			
			found = 1;
			InvsERBlookup(i, sLSPid, tmp, tmp);
			
			if ( strlen(SplicedLSPid) > 0 )
				sprintf(SplicedLSPid,"%s_%d", SplicedLSPid, sLSPid);
			else
				sprintf(SplicedLSPid,"%d", sLSPid);
		}
	
	return (found);
}

void MPLSAddressClassifier::ERBupdate(int entrynb, int LIBptr)
{
	ERB_.Entry_[entrynb].LIBptr_ = LIBptr;
}

int  MPLSAddressClassifier::ERBlocate(int LSPid, int FEC, int &LIBptr)
{
	LIBptr = -1;
	if (LSPid < 0) return(-1);
	for (int i = 0; i < ERB_.NB_; i++)
		if (ERB_.Entry_[i].LSPid_ == LSPid) {
			LIBptr = ERB_.Entry_[i].LIBptr_;
			return(i);
		}
	return -1;
}

int  MPLSAddressClassifier::ERBlocate(int QoSid)
{
	if (QoSid < 0) return(-1);
	
	for (int i=0; i<ERB_.NB_; i++)
		if ( ERB_.Entry_[i].QoSid_ == QoSid )
			return(i);
	
	return(-1);
}

int MPLSAddressClassifier::InvsERBlookup(int LIBptr, int &LSPid, int &FEC, int &QoSid)
{
	LSPid = FEC = QoSid = -1;
	
	if (LIBptr < 0) return(-1);
	
	for (int i=0; i<ERB_.NB_; i++)
		if ( ERB_.Entry_[i].LIBptr_ == LIBptr ) {
			LSPid = ERB_.Entry_[i].LSPid_;
			FEC   = ERB_.Entry_[i].FEC_;
			QoSid = ERB_.Entry_[i].QoSid_;
			return(i);
		}
	
	return(-1);
}

void MPLSAddressClassifier::ERBdump(const char *id)
{
	fflush(stdout);
	printf("      --) ___ERB dump___ [node: %s] (--\n", id);
	printf("--------------------------------------------------------------------------------------\n");
	printf("     FEC     LSPid     LIBptr    SLIBptr    QoSid   aPATHptr  iLabel  iIface  FailNext\n");
	for (int i = 0; i < ERB_.NB_; i++) {
		if (ERB_.Entry_[i].FEC_ == -1) continue;
		
		printf("     %d  ",    ERB_.Entry_[i].FEC_);
		printf("     %d ",     ERB_.Entry_[i].LSPid_);
		printf("     %d     ", ERB_.Entry_[i].LIBptr_);
		printf("     %d  ",    ERB_.Entry_[i].SLIBptr_);
		printf("     %d  ",    ERB_.Entry_[i].QoSid_);
		printf("     %d   ",   ERB_.Entry_[i].aPATHptr_);
		printf("     %d  ",    ERB_.Entry_[i].iLabel_);
		printf("     %d  ",    ERB_.Entry_[i].iIface_);
		printf("     %s\n",    ERB_.Entry_[i].FailNext_);
	}
	printf("\n");
}

//--------------------------------------------------
// LIB (Label Information Base)
//--------------------------------------------------
int MPLSAddressClassifier::LIBnewEntry()
{
	for (int i=0; i<LIB_.NB_; i++) {
		if ( (LIB_.Entry_[i].iIface_ == -1) &&
		     (LIB_.Entry_[i].iLabel_ == -1) &&
		     (LIB_.Entry_[i].oIface_ == -1) &&
		     (LIB_.Entry_[i].oLabel_ == -1) ) {
			LIB_.Entry_[i].LIBptr_ = -1;
			return (i);
		}
	}
	
	if (LIB_.NB_ == MPLS_MaxLIBEntryNB-1) {  
		printf("Error : excess MPLS_MaxLIBEntryNB\n");
		exit(0);
	}
	else {  
		LIB_.NB_++;
		LIBdelete(LIB_.NB_-1);
		return(LIB_.NB_-1);
	}
}

int MPLSAddressClassifier::LIBinsert(int iIface, int iLabel, 
				     int oIface, int oLabel)
{
	int newnb = LIBnewEntry();
	
	if (iIface < 0) iIface = -1;
	if (iLabel < 0) iLabel = -1;
	if (oIface < 0) oIface = -1;
	if (oLabel < 0) oLabel = -1;
	
	LIB_.Entry_[newnb].iIface_ = iIface;
	LIB_.Entry_[newnb].iLabel_ = iLabel;
	LIB_.Entry_[newnb].oIface_ = oIface;
	LIB_.Entry_[newnb].oLabel_ = oLabel;
	
	return(newnb);
}

void MPLSAddressClassifier::LIBdelete(int entrynb)
{
	LIB_.Entry_[entrynb].iIface_ = -1;
	LIB_.Entry_[entrynb].iLabel_ = -1;
	LIB_.Entry_[entrynb].oIface_ = -1;
	LIB_.Entry_[entrynb].oLabel_ = -1;
	LIB_.Entry_[entrynb].LIBptr_ = -1;
	LIB_.Entry_[entrynb].error_  = -1;
}

int MPLSAddressClassifier::LIBisdeleted(int entrynb)
{
	if ((LIB_.Entry_[entrynb].iIface_ == -1) &&
	    (LIB_.Entry_[entrynb].iLabel_ == -1) &&
	    (LIB_.Entry_[entrynb].oIface_ == -1) &&
	    (LIB_.Entry_[entrynb].oLabel_ == -1)) {

		LIB_.Entry_[entrynb].LIBptr_ = -1;
		// delete a entry referencing self in LIB
		for (int i = 0; i < LIB_.NB_; i++)
			if ((LIB_.Entry_[i].LIBptr_ == entrynb))
				LIB_.Entry_[i].LIBptr_ = -1;
		return 0;
	}
	return -1;
}

void MPLSAddressClassifier::LIBupdate(int entrynb, int iIface, int iLabel, 
				      int oIface, int oLabel)
{
	if (iIface != MPLS_DONTCARE)
		LIB_.Entry_[entrynb].iIface_ = iIface;
	if (iLabel != MPLS_DONTCARE)    
		LIB_.Entry_[entrynb].iLabel_ = iLabel;
	if (oIface != MPLS_DONTCARE)
		LIB_.Entry_[entrynb].oIface_ = oIface;
	if (oLabel != MPLS_DONTCARE)
		LIB_.Entry_[entrynb].oLabel_ = oLabel;
}

int MPLSAddressClassifier::LIBlookup(int entrynb, int &oIface, 
				     int &oLabel, int &LIBptr)
{
	oIface = oLabel = LIBptr = -1;

	if (entrynb < 0) return -1;
	
	oIface = LIB_.Entry_[entrynb].oIface_;
	oLabel = LIB_.Entry_[entrynb].oLabel_;
	LIBptr = LIB_.Entry_[entrynb].LIBptr_;
	return 0;
}

int MPLSAddressClassifier::LIBlookup(int iIface, int iLabel, int &oIface, 
				     int &oLabel, int &LIBptr, int &LIBnb)
{
	oIface = oLabel = LIBptr = -1;
	
	if (iLabel < 0) return -1;
	
	for (int i = 0; i < LIB_.NB_; i++)
		if ((LIB_.Entry_[i].iLabel_ == iLabel)) {
			LIBnb = i;
			oIface = LIB_.Entry_[i].oIface_;
			oLabel = LIB_.Entry_[i].oLabel_;
			LIBptr = LIB_.Entry_[i].LIBptr_;
			return 0;
		} 
	return -1;
}

int MPLSAddressClassifier::LIBgetIncoming(int entrynb, int &iIface, 
					  int &iLabel)
{
	if (entrynb < 0)
		return -1;
	iIface = LIB_.Entry_[entrynb].iIface_;
	iLabel = LIB_.Entry_[entrynb].iLabel_;
	return 0;
}

int  MPLSAddressClassifier::InvsLIBlocate(int iLabel, int oLabel, int oIface)
{
	if (oLabel < 0) return(-1);
	
	for (int i=0; i<LIB_.NB_; i++)
		if ( oLabel == 0 ) {
			if ( (LIB_.Entry_[i].iLabel_ == iLabel) &&
			     (LIB_.Entry_[i].oLabel_ == oLabel) &&
			     (LIB_.Entry_[i].oIface_ == oIface))
				return(i);
		}
		else if ( oLabel > 0 ) {
			if ( (LIB_.Entry_[i].oLabel_ == oLabel) &&
			     (LIB_.Entry_[i].oIface_ == oIface))
				return(i);
		}
	
	return(-1);
}

void MPLSAddressClassifier::LIBdump(const char *id)
{
	fflush(stdout);
	printf("    ___LIB dump___ [node: %s]\n", id);
	printf("---------------------------------------------\n");
	printf("       #      iIface     iLabel      oIface     oLabel    LIBptr   Linkerror?\n");
	for (int i = 0; i < LIB_.NB_; i++) {
		if (!LIBisdeleted(i))
			continue;
		printf("     %d:   ", i);
		printf("     %d    ", LIB_.Entry_[i].iIface_);
		printf("     %d    ", LIB_.Entry_[i].iLabel_);
		printf("     %d    ", LIB_.Entry_[i].oIface_);
		printf("     %d    ", LIB_.Entry_[i].oLabel_);
		printf("     %d    ", LIB_.Entry_[i].LIBptr_);
		printf("     %d\n"  , LIB_.Entry_[i].error_);
	}
	printf("\n");
}

//--------------------------------------------------
// MPLS applications
//--------------------------------------------------

int MPLSAddressClassifier::ErLspStacking(int erFEC0,int erLSPid0, 
					 int erFEC, int erLSPid)
{
	int erLIBptr0  =-1, erLIBptr  =-1;

	if ((ERBlocate(erLSPid0,erFEC0,erLIBptr0) < 0) || (erLIBptr0 < 0))
		return -1;
	if ((ERBlocate(erLSPid,erFEC,erLIBptr) < 0) || (erLIBptr < 0))
		return -1;
	LIB_.Entry_[erLIBptr0].LIBptr_ = erLIBptr;
	return 0;
}

int MPLSAddressClassifier::ErLspBinding(int FEC, int PHB, int erFEC, int LSPid)
{
	int LIBptr=-1;
	int erLIBptr=-1;
	
	if ((ERBlocate(LSPid, erFEC, erLIBptr) < 0) ||
	    (erLIBptr < 0)) {
		return -1;
	}
	int PFTnb = PFTlocate(FEC, PHB, LIBptr);
	if ((PFTnb < 0)) {
		PFTinsert(FEC, PHB, erLIBptr);
	}
	else {
		PFTupdate(PFTnb, erLIBptr);
	}
	
	return 0;
}

int MPLSAddressClassifier::FlowAggregation(int fineFEC, int finePHB, 
					   int coarseFEC, int coarsePHB)
{
	int fLIBptr=-1;
	int cLIBptr=-1;

	if ((PFTlocate(coarseFEC,coarsePHB, cLIBptr) < 0) || (cLIBptr < 0))
		return -1;

	int PFTnb = PFTlocate(fineFEC, finePHB, fLIBptr);
	if ((PFTnb < 0))
		PFTinsert(fineFEC, finePHB, cLIBptr);
	else {
		if (fLIBptr < 0)
			PFTupdate(PFTnb, cLIBptr);
		else {
			int i=fLIBptr;
			while (i < 0) {
				fLIBptr = i;;
				i = LIB_.Entry_[fLIBptr].LIBptr_;
			}
			LIB_.Entry_[fLIBptr].LIBptr_ = cLIBptr;
		}
	}
        return 0;
}

int MPLSAddressClassifier::aPathBinding(int FEC, int PHB, int erFEC, int ReLSPid)
{
	int PFTnb, ReERBnb, LIBptr, tmp;
	
	PFTnb   = PFTlocate(FEC, PHB, LIBptr);
	ReERBnb = ERBlocate(ReLSPid, erFEC, tmp);
	
	if ( (PFTnb < 0) || (ReERBnb < 0) ) return(-1);
	
	PFT_.Entry_[PFTnb].aPATHptr_ = ReERBnb;
	
	int ERBnb = InvsERBlookup(LIBptr, tmp, tmp, tmp);
	
	if ( ERBnb > -1 ) {
		ERB_.Entry_[ReERBnb].iLabel_ = ERB_.Entry_[ERBnb].iLabel_;
		ERB_.Entry_[ReERBnb].iIface_ = ERB_.Entry_[ERBnb].iIface_;
	}
	
	return(0);
}

int MPLSAddressClassifier::aPathBinding(int LSPid, int ReLSPid)
{
	int ERBnb, ReERBnb, tmp;
	
	ERBnb   = ERBlocate(LSPid  , -1, tmp);
	ReERBnb = ERBlocate(ReLSPid, -1, tmp);
	
	if ( (ERBnb < 0) || (ReERBnb < 0) ) return(-1);
	
	ERB_.Entry_[ERBnb].aPATHptr_ = ReERBnb;
	ERB_.Entry_[ReERBnb].iLabel_ = ERB_.Entry_[ERBnb].iLabel_;
	ERB_.Entry_[ReERBnb].iIface_ = ERB_.Entry_[ERBnb].iIface_;
	
	return(0);
}

int MPLSAddressClassifier::aPathLookup(int FEC, int PHB, int &LIBnb,
				       int &oIface, int &oLabel, int &LIBptr)
{
	oIface = oLabel = LIBptr = -1;
	
	if (PHB == 0) PHB = -1;
	
	if ( (FEC < 0) || (LIBnb < -1) ) return(-1);
	
	int tmp, ERBnb;
	
	// reroute in flow level
	int PFTnb   = PFTlocate(FEC, PHB, tmp);
	if (PFTnb < 0) {
		PHB = -1;
		PFTnb = PFTlocate(FEC, PHB, tmp);
	}
	
	if (PFTnb < 0) ERBnb = -1; 
	else           ERBnb = PFT_.Entry_[PFTnb].aPATHptr_;
	
	if (ERBnb < 0) {
		// reroute in LSP level
		ERBnb = InvsERBlookup(LIBnb, tmp, tmp, tmp);
		if (ERBnb < 0) return(-1);
		ERBnb = ERB_.Entry_[ERBnb].aPATHptr_;
		if (ERBnb < 0) return(-1);
	}
	if (ERB_.Entry_[ERBnb].iLabel_ > -1) {
		oIface = ERB_.Entry_[ERBnb].iIface_;
		oLabel = ERB_.Entry_[ERBnb].iLabel_;
		LIBptr = ERB_.Entry_[ERBnb].LIBptr_;
		return(0);
	}
	else {
		LIBnb = ERB_.Entry_[ERBnb].LIBptr_;
		return (LIBlookup(ERB_.Entry_[ERBnb].LIBptr_, oIface, oLabel, LIBptr));
	}
}

int MPLSAddressClassifier::SecLSPBinding(int FEC, int PHB, int erFEC, int ReLSPid)
{
	int PFTnb, ReERBnb, LIBptr, tmp;	
	
	PFTnb   = PFTlocate(FEC, PHB, LIBptr);
	ReERBnb = ERBlocate(ReLSPid, erFEC, tmp);
	
	if ( (PFTnb < 0) || (ReERBnb < 0) ) return(-1);
	
	PFT_.Entry_[PFTnb].SLIBptr_ = ReERBnb;
	
	int ERBnb = InvsERBlookup(LIBptr, tmp, tmp, tmp);
	
	if ( ERBnb > -1 ) {
		ERB_.Entry_[ReERBnb].aPATHptr_ = ERB_.Entry_[ERBnb].aPATHptr_;
	}
	
	return(0);
}

int MPLSAddressClassifier::SecLSPBinding(int LSPid, int ReLSPid)
{
	int ERBnb, ReERBnb, tmp;	
	
	ERBnb   = ERBlocate(LSPid  ,-1, tmp);
	ReERBnb = ERBlocate(ReLSPid, -1, tmp);
	
	if ( (ERBnb < 0) || (ReERBnb < 0) ) return(-1);
	
	ERB_.Entry_[ERBnb].SLIBptr_ = ReERBnb;
	
	ERB_.Entry_[ReERBnb].aPATHptr_ = ERB_.Entry_[ERBnb].aPATHptr_;
	
	return(0);
}

int MPLSAddressClassifier::SecLSPLookup(int FEC,int PHB,int &LIBnb,int &oIface,int &oLabel,int &LIBptr)
{
	oIface = oLabel = LIBptr = -1;
	if (PHB ==0)   PHB = -1;
	
	if ( (FEC < 0) || (LIBnb < -1) ) return(-1);
	
	int tmp, ERBnb;
	
	// reroute in flow level
	int PFTnb = PFTlocate(FEC,PHB,tmp);
	
	if ( (PFTnb >= 0) &&
	     (PFT_.Entry_[PFTnb].SLIBptr_ > -1) )
		ERBnb = InvsERBlookup(PFT_.Entry_[PFTnb].SLIBptr_, tmp, tmp, tmp);
	else
		ERBnb = -1;
	
	if ( ERBnb == -1 ) {
		// reroute in LSP level
		ERBnb = InvsERBlookup(LIBnb, tmp, tmp, tmp);
		if (ERBnb < 0) return(-1);
		
		ERBnb = ERB_.Entry_[ERBnb].SLIBptr_;
		if (ERBnb < 0)   return(-1);
	}
	
	if (ERB_.Entry_[ERBnb].iLabel_ > -1) {
		oIface = ERB_.Entry_[ERBnb].iIface_;
		oLabel = ERB_.Entry_[ERBnb].iLabel_;
		LIBptr = ERB_.Entry_[ERBnb].LIBptr_;
		
		return (0);
	}
	else {
		LIBnb = ERB_.Entry_[ERBnb].LIBptr_;
		return (LIBlookup(ERB_.Entry_[ERBnb].LIBptr_, oIface, oLabel, LIBptr));
	}
}

void MPLSAddressClassifier::trace(char *ptype, int psize, int ilabel, 
				  char *op, int oiface, int olabel, int ttl)
{
	if (trace_mpls_ != 1) return;
	
	Tcl::instance().evalf("%s trace-packet-switching " TIME_FORMAT 
			      " %d %d %s %d %s %d %d %d %d",
			      name(),
			      Scheduler::instance().clock(),
			      PI_.srcnode_,
			      PI_.dst_.addr_,
			      ptype,
			      ilabel,
			      op,
			      oiface,
			      olabel,
			      ttl,
			      psize);
}
