###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################
###########################################################################
#         < An example of MPLS Simulator >                                #
#                                                                         #
#     File: an example for simulating resource preemption in CR-LSP       #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Aug. 2000                    #
###########################################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

set ns [new Simulator]

set na [open test-crlsp-pmt.tr w]
#set nf [open test-crlsp-pmt.nam w]
set fSBT [open SBT w]
set fHBT [open HBT w]
set fRT1 [open RT1 w]
set fRT2 [open RT2 w]

set TpktSBT 0
set TpktHBT 0
set TpktRT1 0
set TpktRT2 0
proc record {} {
        global fSBT fHBT fRT1 fRT2 SBTsink HBTsink RT1sink RT2sink TpktSBT TpktHBT TpktRT1 TpktRT2

	set ns [Simulator instance]
	
	#Set the time after which the procedure should be called again
        set time 1.0
	#set time 0.045
	
	#How many bytes have been received by the traffic sink?
        set bwSBT [$SBTsink set bytes_]
        set bwHBT [$HBTsink set bytes_]
        set bwRT1 [$RT1sink set bytes_]
        set bwRT2 [$RT2sink set bytes_]
	
	#Get the current time
        set now [$ns now]

	#Calculate the bandwidth (in MBit/s) and write it to the file
        puts $fSBT "$now [expr $bwSBT/$time*8/1000000]"
        puts $fHBT "$now [expr $bwHBT/$time*8/1000000]"
        puts $fRT1 "$now [expr $bwRT1/$time*8/1000000]"
        puts $fRT2 "$now [expr $bwRT2/$time*8/1000000]"
        

	#Reset the bytes_ values on the traffic sink
        $SBTsink set bytes_ 0
        $HBTsink set bytes_ 0
        $RT1sink set bytes_ 0
        $RT2sink set bytes_ 0
        
        set pktSBT  [expr $bwSBT / 200]
        set TpktSBT [expr $TpktSBT + $pktSBT]
        set pktHBT  [expr $bwHBT / 200]
        set TpktHBT [expr $TpktHBT + $pktHBT]
        set pktRT1  [expr $bwRT1 / 200]
        set TpktRT1 [expr $TpktRT1 + $pktRT1]
        set pktRT2  [expr $bwRT2 / 200]
        set TpktRT2 [expr $TpktRT2 + $pktRT2]

	#Re-schedule the procedure
        $ns at [expr $now+$time] "record"
}

proc recv-pkts {} {
     global TpktSBT TpktHBT TpktRT1 TpktRT2

     puts "The packet number of SBT : $TpktSBT"
     puts "The packet number of HBT : $TpktHBT"
     puts "The packet number of RT1 : $TpktRT1"
     puts "The packet number of RT2 : $TpktRT2"
}


proc attach-expoo-traffic { node sink size burst idle rate } {
        global ns

	set source [new Agent/CBR/UDP]
	$ns attach-agent $node $source

	set traffic [new Traffic/Expoo]
	$traffic set packet-size $size
	$traffic set burst-time $burst
	$traffic set idle-time $idle
	$traffic set rate $rate

	$source attach-traffic $traffic

	$ns connect $source $sink
	return $source
}


$ns trace-all $na
#$ns namtrace-all $nf
proc finish {} {
	global ns na nf fSBT fHBT fRT1 fRT2
	$ns flush-trace
        close $na
	#close $nf
	close $fSBT
	close $fHBT
	close $fRT1
	close $fRT2
	
	recv-pkts
#        exec nam  test-crlsp-pmt.nam &
        #exec xgraph -m SBT HBT RT1 RT2 -geometry 800x400 &
	exit 0
}

#
# make nodes & MPLSnodes
#
set Node0 [$ns node]
set LSR1  [$ns mpls-node]
set LSR2  [$ns mpls-node]
set LSR3  [$ns mpls-node]
set Node4 [$ns node]


#
# make links
#
$ns duplex-link $Node0  $LSR1  2Mb  10ms DropTail
$ns duplex-link $LSR1   $LSR2  1Mb  10ms CBQ
$ns duplex-link $LSR2   $LSR3  1Mb  10ms CBQ
$ns duplex-link $LSR3   $Node4 1Mb  10ms DropTail


$ns duplex-link-op $LSR1 $LSR2 queuePos 0.8

#
# configure ldp agents on all mpls nodes
#
$ns configure-ldp-on-all-mpls-nodes


# configure-cbq-for-SBTS {qlim cbq_qtype okborrow bw maxidle extradelay}
$ns cfg-cbq-for-SBTS      10  DropTail    1     0.14  auto       0
$ns cfg-cbq-for-HBTS      10  DropTail    1     0.01  auto       0
$ns cfg-cbq-for-RTS       10  DropTail    0     0.8   auto       0
$ns cfg-cbq-for-STS       10  DropTail    1     0.05  auto       0

$ns bind-flowid-to-SBTS 100
$ns bind-flowid-to-SBTS 200 200
$ns bind-flowid-to-SBTS 300 300 
$ns bind-flowid-to-SBTS 400 400

$ns bind-ldp-to-STS

#
# set ldp-message clolr
#
$ns ldp-request-color       blue
$ns ldp-mapping-color       red
$ns ldp-withdraw-color      magenta
$ns ldp-release-color       orange
$ns ldp-notification-color  yellow

#$LSR2 trace-mpls

# set ldp events
#
#
# $ns enable-control-driven
#

#$ns enable-data-driven
#$ns enable-on-demand
#$ns enable-ordered-control

#
#Create a traffic sink and attach it to the node Node4
#

set SBTsink [new Agent/LossMonitor]
$ns attach-agent $Node4  $SBTsink
$SBTsink clear

set HBTsink [new Agent/LossMonitor]
$ns attach-agent $Node4  $HBTsink
$HBTsink clear

set RT1sink [new Agent/LossMonitor]
$ns attach-agent $Node4  $RT1sink
$RT1sink clear

set RT2sink [new Agent/LossMonitor]
$ns attach-agent $Node4  $RT2sink
$RT2sink clear

#
#Create a traffic source
#

set SBT [attach-expoo-traffic $Node0  $SBTsink 200 0 0  250k]
$SBT set fid_ 100
$ns color 100 red

set HBT [attach-expoo-traffic $Node0  $HBTsink 200 0 0  250k]
$HBT set fid_ 200
$ns color 200 blue

set RT1 [attach-expoo-traffic $Node0  $RT1sink 200 0 0  600k]
$RT1 set fid_ 300
$ns color 300 yellow

set RT2 [attach-expoo-traffic $Node0  $RT2sink 200 0 0  700k]
$RT2 set fid_ 400
$ns color 400 black


proc notify-erlsp-setup {node lspid} {
        
        set ns [Simulator instance]
        puts "[$ns now] nodeid=[$node id] : lspid=$lspid"
        
        set module [$node get-module "MPLS"]
	
        if {[$node id] == 1} {
           
               switch $lspid {
                      1000 {  $module bind-flow-erlsp   4  100  $lspid
                           }
                      1100 {  $module bind-flow-erlsp   4  200  $lspid 
                              $module set-flow-prio     4  200  1
                           }
                      1200 { $module bind-flow-erlsp   4  300  $lspid }
                      1300 { $module bind-flow-erlsp   4  400  $lspid }
                 default { 
                           puts "error" 
                           exit 1
                         }
               }
        }
        
}

proc notify-erlsp-fail {node status lspid tr} {

   puts "nodeid=[$node id] : status=$status  lspid=$lspid  tr=$tr"

}

#
# set MPLS modules
#
for {set i 1} {$i < 4} {incr i} {
	set a LSR$i
	set m [eval $$a get-module "MPLS"]
	eval set LSRmpls$i $m
}

$ns at 0.0  "record"

$ns at 0.1 "$LSRmpls1 setup-erlsp  3  1_2_3  1000"
$ns at 0.1 "$LSRmpls1 setup-erlsp  3  1_2_3  1100"

############################################################################
# MplsNode instproc setup-crlsp {fec er lspid TRate BSize PSize SPrio HPrio}
#
$ns at 0.1  "$LSRmpls1 setup-crlsp  3  1_2_3  1200  600K  400B 200B 7 4"
############################################################################

$ns at 1.0  "$SBT start"
$ns at 1.0  "$HBT start"
$ns at 1.0  "$RT1 start"


############################################################################
# MplsNode instproc setup-crlsp {fec er lspid TRate BSize PSize SPrio HPrio}
#
$ns at 10.0  "$LSRmpls1 setup-crlsp  3  1_2_3  1300  700K  400B 200B 3 2"
############################################################################

$ns at 11.0  "$RT2 start"

$ns at 30.0  "$RT2 stop"
$ns at 31.0  "$LSRmpls1 send-crldp-release-msg  1300"

$ns at 40.0  "$SBT stop"
$ns at 40.0  "$HBT stop"
$ns at 40.0  "$RT1 stop"

$ns at 43.0 "finish"

$ns run

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
