###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################
###########################################################################
#         < An example of MPLS Simulator >                                #
#                                                                         #
#     File: an example for simulating basic MPLS function                 #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Jan. 2000                    #
###########################################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

set ns [new Simulator]

set na [open test-lsp.tr w]
$ns trace-all $na
set nf [open test-lsp.nam w]
$ns namtrace-all $nf
proc finish {} {
	global ns na nf
	$ns flush-trace
	close $na
	close $nf
        #exec nam test-lsp.nam &
	exit 0
}


#
# make nodes & MPLSnodes
#
set Node0  [$ns node]
set Node1  [$ns node]
set LSR2   [$ns mpls-node]
set LSR3   [$ns mpls-node]
set LSR4   [$ns mpls-node]
set LSR5   [$ns mpls-node]
set LSR6   [$ns mpls-node]
set LSR7   [$ns mpls-node]
set LSR8   [$ns mpls-node]
set Node9  [$ns node]
set Node10 [$ns node]


#
# make links
#
$ns duplex-link $Node0 $LSR2   1Mb  10ms DropTail
$ns duplex-link $Node1 $LSR2   1Mb  10ms DropTail

$ns duplex-link $LSR2  $LSR3   1Mb  10ms DropTail
$ns duplex-link $LSR3  $LSR4   1Mb  10ms DropTail
$ns duplex-link $LSR4  $LSR8   1Mb  10ms DropTail

$ns duplex-link $LSR2  $LSR5   1Mb  10ms DropTail
$ns duplex-link $LSR5  $LSR6   1Mb  10ms DropTail
$ns duplex-link $LSR5  $LSR4   1Mb  10ms DropTail
$ns duplex-link $LSR6  $LSR7   1Mb  10ms DropTail
$ns duplex-link $LSR6  $LSR8   1Mb  10ms DropTail
$ns duplex-link $LSR7  $LSR8   1Mb  10ms DropTail

$ns duplex-link $LSR7  $Node9  1Mb  10ms DropTail
$ns duplex-link $LSR8  $Node10 1Mb  10ms DropTail


#
# configure ldp agents on all mpls nodes
#
$ns configure-ldp-on-all-mpls-nodes


#
# set ldp-message colors
#
$ns ldp-request-color       blue
$ns ldp-mapping-color       red
$ns ldp-withdraw-color      magenta
$ns ldp-release-color       orange
$ns ldp-notification-color  yellow


#
# set ldp events
#
$ns enable-control-driven
#$ns enable-data-driven
#$ns enable-on-demand
#$ns enable-ordered-control


#
# trace mpls packets
#
#$LSR1  trace-mpls
#$LSR2  trace-mpls

#$LSR1 set ttl_ 32


#
# create traffic generators
#
set Src0 [new Application/Traffic/CBR]
$Src0 set packetSize_ 500
$Src0 set interval_ 0.010

set Src1 [new Application/Traffic/CBR]
$Src1 set packetSize_ 500
$Src1 set interval_ 0.010


#
# create UDP agents
#
set Udp0 [new Agent/UDP]
set Udp1 [new Agent/UDP]

$Src0 attach-agent $Udp0
$Src1 attach-agent $Udp1

$ns attach-agent $Node0 $Udp0
$ns attach-agent $Node1 $Udp1


set Dst0 [new Agent/Null]
$ns attach-agent $Node9 $Dst0

set Dst1 [new Agent/Null]
$ns attach-agent $Node10 $Dst1

$ns connect $Udp0 $Dst0
$ns connect $Udp1 $Dst1

$ns at 0.1  "$Src0 start"
$ns at 0.1  "$Src1 start"


#
# set MPLS modules
#
for {set i 2} {$i < 9} {incr i} {
	set a LSR$i
	set m [eval $$a get-module "MPLS"]
	eval set LSRmpls$i $m
}


#
# set simulation events
#
$ns at 0.2 "$LSRmpls7 send-ldp-withdraw-msg  9"
$ns at 0.2 "$LSRmpls8 send-ldp-withdraw-msg 10"

$ns at 0.3 "$LSRmpls2 aggregate-flows    9  6 "
$ns at 0.3 "$LSRmpls2 aggregate-flows   10  6 "
$ns at 0.5 "$LSRmpls6 send-ldp-withdraw-msg 6"
$ns at 0.7 "$Src1 stop"

$ns at 0.7 "$LSRmpls2 setup-erlsp  7  5_4_8_6_7  3000"
$ns at 0.9 "$LSRmpls2 bind-flow-erlsp  9 -1      3000"
$ns at 1.1 "$LSRmpls2 send-crldp-release-msg     3000"

$ns at 1.2 "$LSRmpls4 setup-erlsp     8  4_5_6_8        3500"
$ns at 1.4 "$LSRmpls2 setup-erlsp     7  2_3_4_L3500_7  3600"
$ns at 1.6 "$LSRmpls2 bind-flow-erlsp  9 -1             3600"

$ns at 1.7 "$LSRmpls2 pft-dump"
$ns at 1.7 "$LSRmpls2 erb-dump"
$ns at 1.7 "$LSRmpls2 lib-dump"

$ns at 1.7 "$LSRmpls4 pft-dump"
$ns at 1.7 "$LSRmpls4 erb-dump"
$ns at 1.7 "$LSRmpls4 lib-dump"

$ns at 1.8 "$LSRmpls7 send-crldp-withdraw-msg  3600"
$ns at 1.9 "$LSRmpls8 send-crldp-withdraw-msg  3500"

$ns at 2.0 "$Src0 stop"

$ns at 3.0 "finish"

$ns run

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
