###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################
###########################################################################
#         < An example of MPLS Simulator >                                #
#                                                                         #
#     File: An example for path restoration simulation                    #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Aug. 2000                    #
###########################################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

proc usage { } {
	puts stderr {usage: ns test-reroute.tcl reroute-option
The reroute-options are as follows: 
    - drop             : drop all traffic around link failure.
    - L3data-driven    : using L3 routing protocol and LDP data-driven trigger.
    - L3control-driven : using L3 routing protocol and LDP control-driven trigger
    - haskin           : the scheme proposed by haskin.
    - makam            : the scheme proposed by makam.
    - simple-dynamic   : establishing new backup LSP between the node that detected a failure and PML.
    - shortest-dynamic : establishing new backup LSP between the node that detected a failure and the next node.
       }
}

set ns [new Simulator]

set na [open test-reroute.tr w]
set nf [open test-reroute.nam w]
$ns trace-all $na
$ns namtrace-all $nf
set f0 [open test-reroute-bw.tr w]
set fs [open test-reroute-seq.tr w]
proc finish {} {
	global ns na nf f0 fs
	$ns flush-trace
	close $na
	close $nf
	close $f0
	close $fs
        #exec nam test-reroute.nam &
        #exec xgraph -m test-reroute.tr -geometry 800x400 &
	#exec xgraph -m -nl test-reroute-seq.tr -geometry 800x400 &
	exit 0
}

proc attach-expoo-traffic {node sink size burst idle rate} {
        global ns
	
	set source [new Agent/CBR/UDP]
	$ns attach-agent $node $source
	
	set traffic [new Traffic/Expoo]
	$traffic set packet-size $size
	$traffic set burst-time $burst
	$traffic set idle-time $idle
	$traffic set rate $rate
	
	$source attach-traffic $traffic
	
	$ns connect $source $sink
	return $source
}

# Define a procedure which periodically records the bandwidth received by the
# traffic sink sink0 and writes it to the file f0.
set totalpkt 0
proc record {} {
        global sink0 f0 totalpkt
	
	set ns [Simulator instance]
	
	#Set the time after which the procedure should be called again
        set time 0.065
	
	#How many bytes have been received by the traffic sink?
        set bw0 [$sink0 set bytes_]
	
	#Get the current time
        set now [$ns now]
	
	#Calculate the bandwidth (in MBit/s) and write it to the file
        puts $f0 "$now [expr $bw0/$time*8/1000000]"
	
	#Reset the bytes_ values on the traffic sink
        $sink0 set bytes_ 0
	
	#Re-schedule the procedure
        $ns at [expr $now+$time] "record"
        
        set bw0 [expr $bw0 / 200]
        set totalpkt [expr $totalpkt + $bw0]       
}

proc recv-pkts {} {
	global totalpkt seqerrnb
	
	puts "The Number of Total received packet is $totalpkt"
	puts "The Number of Total unordered packet is $seqerrnb"
}

set prvseqnb -1
set seqerrnb 0
proc seq-record {size rate ftime} {
        global prvseqnb seqerrnb sink0 fs 
	
	set ns [Simulator instance]
	
	#Set the time after which the procedure should be called again
        set tsize [parse-bw $size]
        set trate [parse-bw $rate]
        set time [expr double($tsize)/double($trate)/8.0]
	
	#Get the current time
        set now [$ns now]
	
	# seek the sequence number of packet.
	set revseqnb [$sink0 set expected_]
	
	if {$prvseqnb > $revseqnb} {
		incr seqerrnb 1
	}
	
	# write the sequence number of packet to the file
        if {$prvseqnb != $revseqnb} {
		puts $fs "$now [$sink0 set expected_]"
		set prvseqnb $revseqnb
        }
	
	#Re-schedule the procedure
        if { [expr $now+$time] < $ftime } {
		$ns at [expr $now+$time] "seq-record $size $rate $ftime"
        }
}

# routing protocol
$ns rtproto DV

#
# make nodes & MPLSnodes
#
set node0  [$ns node]
set LSR1   [$ns mpls-node]
set LSR2   [$ns mpls-node]
set LSR3   [$ns mpls-node]
set LSR4   [$ns mpls-node]
set LSR5   [$ns mpls-node]
set LSR6   [$ns mpls-node]
set LSR7   [$ns mpls-node]
set LSR8   [$ns mpls-node]
set LSR9   [$ns mpls-node]
set node10 [$ns node]

#
# make links
#
$ns duplex-link $node0 $LSR1  1Mb  10ms DropTail
$ns duplex-link $LSR1  $LSR3  1Mb  10ms DropTail
$ns duplex-link $LSR3  $LSR5  1Mb  10ms DropTail
$ns duplex-link $LSR5  $LSR7  1Mb  10ms DropTail
$ns duplex-link $LSR7  $LSR9  1Mb  10ms DropTail

$ns duplex-link $LSR1  $LSR2  1Mb  10ms DropTail
$ns duplex-link $LSR2  $LSR4  1Mb  10ms DropTail
$ns duplex-link $LSR4  $LSR6  1Mb  10ms DropTail
$ns duplex-link $LSR6  $LSR8  1Mb  10ms DropTail
$ns duplex-link $LSR8  $LSR9  1Mb  10ms DropTail 

$ns duplex-link $LSR1  $LSR2  1Mb  10ms DropTail
$ns duplex-link $LSR3  $LSR4  1Mb  10ms DropTail
$ns duplex-link $LSR5  $LSR6  1Mb  10ms DropTail
$ns duplex-link $LSR7  $LSR8  1Mb  10ms DropTail

$ns duplex-link $LSR9  $node10  1Mb  10ms DropTail

#
# configure ldp agents on all mpls nodes
#
$ns configure-ldp-on-all-mpls-nodes

#
# set ldp-message clolr
#
$ns ldp-request-color       blue
$ns ldp-mapping-color       red
$ns ldp-withdraw-color      magenta
$ns ldp-release-color       orange
$ns ldp-notification-color  green

proc set-reroute-init {option} {
	global ns LSR1 LSR9 LSR5
	
	set LSRmpls1 [$LSR1 get-module "MPLS"]
	set LSRmpls5 [$LSR5 get-module "MPLS"]
	set LSRmpls9 [$LSR9 get-module "MPLS"]
	
        switch $option {
            drop                 { # set reroute action to take when a link fails
                                   $ns enable-reroute drop
                                   $LSRmpls1 enable-data-driven
                                 }
            L3data-driven        { # set reroute action to take when a link fails
                                   $ns enable-reroute L3
                                   $LSRmpls5 set-protection-flow 0.1 0.01 10 -1
                                   
                                   # set ldp events
                                   $ns enable-on-demand
                                   $ns enable-ordered-control
                                   $ns enable-data-driven
                                 }
            L3control-driven     { 
                                   # set reroute action to take when a link fails
                                   $ns enable-reroute drop
                                   
                                   # set ldp events
                                   $ns enable-control-driven
                                 }
            simple-dynamic       { # set reroute action to take when a link fails
                                   $ns enable-reroute simple-dynamic
                                   $LSRmpls5 set-protection-flow 0.1 0.01 10 -1

                                   # set ldp events
                                   $LSRmpls1 enable-data-driven
                                 }
            haskin               { # set reroute action to take when a link fails
                                   $ns enable-reroute drop
                                   $ns set-protection-lsp 0.7 0.01 1000
                                 }
            makam                { # set reroute action to take when a link fails
                                   $ns enable-reroute notify-prenegotiated
                                   $LSRmpls5 set-protection-lsp 0.7 0.01 1000
                                 }
            shortest-dynamic     { # set reroute action to take when a link fails
                                   $ns enable-reroute shortest-dynamic
                                   $LSRmpls5 set-protection-lsp 0.7 0.01 1000
                                   $LSRmpls9 enable-reroute-egress-lsr
                                 }
            default              { usage
                                   exit
                                 }
        }
}

proc set-reroute-event {option} {
	global ns LSR1 LSR3 LSR5 LSR7 LSR9
	
	set LSRmpls1 [$LSR1 get-module "MPLS"]
	set LSRmpls3 [$LSR3 get-module "MPLS"]
	set LSRmpls5 [$LSR5 get-module "MPLS"]
	set LSRmpls7 [$LSR7 get-module "MPLS"]
	set LSRmpls9 [$LSR9 get-module "MPLS"]
	
        switch $option {
            haskin               { 
                                   # The setup of working LSP
                                   $ns at 0.0  "$LSRmpls1  setup-erlsp  9   3_5_7_9     1000"
                            
                                   # The setup of alternative LSP
                                   $ns at 0.1  "$LSRmpls1  setup-erlsp  9   2_4_6_8_9      2000"
                                   $ns at 0.2  "$LSRmpls9  setup-erlsp  9   7_5_3_1_L2000  2005"

                                   # bind a flow to LSP
                                   $ns at 0.3  "$LSRmpls1 bind-flow-erlsp   10   100   1000"

                                   # binding working LSP to alternative LSP
                                   $ns at 0.3  "$LSRmpls1 reroute-lsp-binding      1000    2000"
                                   $ns at 0.3  "$LSRmpls3 reroute-lsp-binding      1000    2005"
                                   $ns at 0.3  "$LSRmpls5 reroute-lsp-binding      1000    2005"
                                   $ns at 0.3  "$LSRmpls7 reroute-lsp-binding      1000    2005"
                                 }
            makam                { 
                                   # The setup of working LSP
                                   $ns at 0.0  "$LSRmpls1  setup-erlsp  9   3_5_7_9     1000"
                            
                                   # The setup of alternative LSP
                                   $ns at 0.1  "$LSRmpls1  setup-erlsp  9   2_4_6_8_9      2000"

                                   # bind a flow to LSP
                                   $ns at 0.3  "$LSRmpls1 bind-flow-erlsp   10   100   1000"

                                   # binding working LSP to alternative LSP
                                   $ns at 0.3  "$LSRmpls1 reroute-lsp-binding      1000    2000"
                                 }
            shortest-dynamic     { 
                                   # The setup of working LSP
                                   $ns at 0.0  "$LSRmpls1  setup-erlsp  9   3_5_7_9     1000"
                                   
                                   # bind a flow to LSP
                                   $ns at 0.3  "$LSRmpls1 bind-flow-erlsp   10   100   1000"
                                 }
        }
}

# Create a traffic sink and attach it to the node node9
set sink0 [new Agent/LossMonitor]
$ns attach-agent $node10 $sink0
$sink0 clear

# Create a traffic source
set src0 [attach-expoo-traffic $node0 $sink0 200 0 0 500k]
$src0 set fid_ 100
$ns color 100 magenta

proc notify-erlsp-setup {node lspid} {
        set ns [Simulator instance]
	set module [$node get-module "MPLS"]
	
        if { $lspid==1001 } {
		$module secondary-lsp-binding 1000 1001
        }
}

proc notify-erlsp-fail {node status lspid tr} {
        set ns [Simulator instance]
	set module [$node get-module "MPLS"]
        
        if { [$node id] == 1 && $status=="BSNodeError" } {
		$module set-lib-error-for-lspid $lspid 1
        }
        if { [$node id] == 1 && $status=="NodeRepair" } {
		$module set-lib-error-for-lspid $lspid -1
        }        
}

# reroute mechanisms
if {$argc == 1} {
	set-reroute-init  $argv
	set-reroute-event $argv
} else {
	usage
	exit
}

$ns at 0.0 "record"
$ns at 0.3 "seq-record 200 500k 2.0"

$ns at 0.3 "$src0 start"

$ns rtmodel-at 0.8 down $LSR5 $LSR7
$ns rtmodel-at 1.3 up   $LSR5 $LSR7

$ns at 1.8 "$src0 stop"

$ns at 2.0 "recv-pkts"
$ns at 2.0 "record"
$ns at 2.0 "finish"

$ns run

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
