###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                               	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################

#############################################################
#                                                           #
#     File: Constraint-based Routing in MPLS                #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Sep. 2000      #
#                                                           #
#############################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

Simulator instproc build-topology-db {allnodes src fnodes flinks} {
	global TE_DB
        
	set len [llength $allnodes]
	
	for { set i 0 } { $i < $len } { incr i 1 } {
		set me       [lindex $allnodes $i]
		set neighbor [$me set neighbor_]
		
		for { set j 0 } { $j < $len } { incr j 1 } {
			
			set TE_DB($i,$j) 1000
			set you [lindex $allnodes $j]
			
			if { $i == $j } {
				set TE_DB($i,$j) 0
			} elseif { [lsearch $fnodes $me] > -1 || [lsearch $fnodes $you] > -1 } {
				# node failure
				continue
			} else {
				
				set ix [lsearch $flinks $me]
				if { $ix > -1 && [lindex $flinks [expr $ix +1]] == $you } {
					# link failure
					continue
				}
				
				set is_n [lsearch $neighbor $you]
				if { $is_n >= 0 } { 
					# $you is neighbor of $me
					# sigh..
					# This is not exact link information.
					# This code must be updated.....
					
					set diff [expr abs([$src get-rtmetric $me]-[$src get-rtmetric $you])]
					if { $diff == 1 } {
						set TE_DB($i,$j) 1
					}
				}
			}
		}
	}
}

Simulator instproc shortest-path { len source dstid } {
	global TE_DB
	
	for { set i 0 } { $i < $len } { incr i 1 } {
		set found($i) 0
		set distance($i) $TE_DB($source,$i)
		if { $distance($i) < 1000 } {
			set path($i)  $i
		} else {
			set path($i) -1
		}
	}
	
	for { set i 0 } { $i < $len } { incr i } {
		set min 1000
		set minpos -1 
		for { set j 0 } { $j < $len } { incr j 1 } {
			if { $found($j) == 0 && $distance($j) < $min } {
				set min $distance($j)
				set minpos $j
			}
		}
		
		if {$minpos == -1} {
			continue
		}
		
		set found($minpos) 1
		
		for { set w 0 } { $w < $len } { incr w 1 } {
			if { $found($w) == 0 } {
				if { [expr $distance($minpos) + $TE_DB($minpos,$w)] < $distance($w) } {
					set distance($w) [expr $distance($minpos) + $TE_DB($minpos,$w)]
					set path($w) $minpos
					set numw $w
				}
			}
		}
	}
	
	set way ""
	set myway ""
	
	while { $path($dstid) != $dstid } { 
		if { $path($dstid) == -1 } {
			# puts "No Path"
			return -1
		}
		lappend way $path($dstid)
		set dstid $path($dstid)
	}
	
	set a [llength $way]
	for { set i 0 } { $i < $a } { incr i 1 } { 
		set temp [lindex $way [expr $a - $i - 1]]
		lappend myway [expr $temp]
	}
	return $myway
}

Simulator instproc mpls-node-addr { id } {    
	set allnodes [$self set linked_mplsnodes_]
	
	set len [llength $allnodes]
	for { set i 0 } { $i < $len } { incr i 1 } { 
		set lsr [lindex $allnodes $i]
		if { [$lsr id] == $id } {
			return $lsr
		}
	}
	puts "mpld-node-addr: no object"
	exit
}

Simulator instproc get-shortest-path {src dst fnodeids flinkids} {
	global TE_DB
	
	set allnodes [$self set linked_mplsnodes_]
	
	set src [$self mpls-node-addr $src]
	set dst [$self mpls-node-addr $dst]
	
	set fnodes ""
	set flinks ""
	
	set len [llength $fnodeids]
	for { set i 0 } { $i < $len } { incr i 1 } { 
		set lsr [$self mpls-node-addr [lindex $fnodeids $i]]
		lappend fnodes $lsr
	}
	set len [llength $flinkids]
	for { set i 0 } { $i < $len } { incr i 1 } { 
		set lsr [$self mpls-node-addr [lindex $flinkids $i]]
		lappend flinks $lsr
	}
	
	$self build-topology-db $allnodes $src $fnodes $flinks
	
	set srcix [lsearch $allnodes $src]
	set dstix [lsearch $allnodes $dst]
	
	set len [llength $allnodes]
	set ret [$self shortest-path $len $srcix $dstix]
	
	if {$ret == -1} {
		return $ret
	} else {
		set len [llength $ret]
		append er [$src id]
		
		for { set i 0 } { $i < $len } { incr i 1 } { 
			append er _
			set hopid [[lindex $allnodes [lindex $ret $i]] id]
			append er $hopid
		}
		
		append er _
		append er [$dst id]
		return $er
	}
}

proc build-qos-db { allnodes rbw } {
        global TE_DB
        
	set dbptr [open ./te.db r]
	
	set len [llength $allnodes]
	
	for { set i 0 } { $i < $len } { incr i 1 } {
		
		for { set j 0 } { $j < $len } { incr j 1 } {
			
			gets $dbptr bw
			if { $i != $j && $bw < $rbw } {
				set TE_DB($i,$j) 1000
			} else {
				set TE_DB($i,$j) [expr $bw / 1000000000]
			}
		}
	}
	
	close $dbptr
}

Simulator instproc get-qos-path {allnodes src dst bw} {
	global TE_DB
	
	set bw [parse-bw $bw]
	set len [llength $allnodes]
	for { set i 0 } { $i < $len } { incr i 1 } { 
		set lsr [lindex $allnodes $i]
		if { [$lsr id] == $src } {
			set src $lsr
		} elseif { [$lsr id] == $dst } {
			set dst $lsr
		}
	}
	
	build-qos-db $allnodes $bw
	
	set srcix [lsearch $allnodes $src]
	set dstix [lsearch $allnodes $dst]
	
	set baseroute [$self shortest-path $len $srcix $dstix]
	set hopcount  [llength $baseroute]
	
	set route $baseroute
	set rlen  $hopcount
	
	while {$route != -1} {
		$self update-resource-info $allnodes $route $srcix $dstix $bw
		
		set route [$self shortest-path $len $srcix $dstix]
		set rlen  [llength $route]
		if { $route != -1 && $rlen < $hopcount } {
			set baseroute $route
			set hopcount  $rlen
		}
	}
	
	if {$baseroute == -1} {
		return $baseroute
	} else {
		build-qos-db $allnodes -1
		$self update-resource-info $allnodes $baseroute $srcix $dstix $bw
		$self save-resource-info $allnodes
		
		set len [llength $baseroute]
		append er [$src id]
		
		for { set i 0 } { $i < $hopcount } { incr i 1 } { 
			append er _
			set hopid [[lindex $allnodes [lindex $baseroute $i]] id]
			append er $hopid
		}
		
		append er _
		append er [$dst id]
		return $er
	}
}

Simulator instproc update-resource-info {allnodes route srcix dstix bw} {
	global TE_DB
        
	if { $route == -1} {
		return
	}
	
	set rlen  [llength $route]
	for { set i 0 } { $i <= $rlen } { incr i 1 } { 
		if {$i == 0} {
			set hop1 [lindex $allnodes $srcix]
		} else {
			set hop1 [lindex $allnodes [lindex $route [expr $i -1]]]
		}
		
		if {$i >= $rlen} {
			set hop2 [lindex $allnodes $dstix]
		} else {
			set hop2 [lindex $allnodes [lindex $route $i]]
		}
		
		set hop1ix [lsearch $allnodes $hop1]
		set hop2ix [lsearch $allnodes $hop2]
		
		set TE_DB($hop1ix,$hop2ix) [expr $TE_DB($hop1ix,$hop2ix) * 1000000000 - $bw]
		
		if { $TE_DB($hop1ix,$hop2ix) <= 0 } {
			set TE_DB($hop1ix,$hop2ix) 1000
		} else {
			set TE_DB($hop1ix,$hop2ix) [expr $TE_DB($hop1ix,$hop2ix) / 1000000000]
		}
	}
}

Simulator instproc save-resource-info {allnodes} {
	global TE_DB
	
	set dbptr [open ./te.db w]
	
	set len [llength $allnodes]
	for { set i 0 } { $i < $len } { incr i 1 } {
		for { set j 0 } { $j < $len } { incr j 1 } {
			if { $TE_DB($i,$j) == 1000 } {
				puts $dbptr 0
			} else {
				puts $dbptr [expr $TE_DB($i,$j) * 1000000000]
			}
		}
	}
	
	flush $dbptr
	close $dbptr
}

Simulator instproc collect-resource-info {interval} {
	set dbptr [open ./te.db w]
	set allnodes [$self set linked_mplsnodes_]
	set len [llength $allnodes]
	
	for { set i 0 } { $i < $len } { incr i 1 } {
		set me [lindex $allnodes $i]
		set neighbor [$me set neighbor_]
		
		for { set j 0 } { $j < $len } { incr j 1 } {
			set node1 [lindex $allnodes $i]
			set node2 [lindex $allnodes $j]
			set module1 [$node1 get-module "MPLS"]
			set module2 [$node2 get-module "MPLS"]
			
			if { $i == $j } {
				puts $dbptr 0
			} else {
				set nlen [llength $neighbor]
				for { set k 0 } { $k < $nlen } { incr k 1 } {
					set n [lsearch $neighbor [lindex $allnodes $j]]
					
					if { $n >= 0} {
						set linkptr [$module1 get-link-ptr $node2]
						puts $dbptr [$linkptr get-usable-bw]
						break
					}
				}
				if {$k ==$nlen} {
					puts $dbptr 1000
				}
			}
		}
	}
	
	flush $dbptr
	close $dbptr
	
	set now [$self now]
	$self at [expr $now + $interval] "$self collect-resource-info $interval"
}

RtModule/MPLS instproc constraint-based-routing {dst bw} {
	$self instvar classifier_
	
	if { [$classifier_ set is_cbq_installed_] != 1 } {
		exit
	}
	set ns [Simulator instance]
	set allnodes [$ns set linked_mplsnodes_]
	set er [$ns get-qos-path $allnodes [[$self node] id] $dst $bw]
	
	return $er
}

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
