# -*-	Mode:tcl; tcl-indent-level:8; tab-width:8; indent-tabs-mode:t -*-
#
# Time-stamp: <2000-09-11 14:54:38 haoboy>
# 
#  Copyright (c) 1997 by the University of Southern California
#  All rights reserved.
# 
#  Permission to use, copy, modify, and distribute this software and its
#  documentation in source and binary forms for non-commercial purposes
#  and without fee is hereby granted, provided that the above copyright
#  notice appear in all copies and that both the copyright notice and
#  this permission notice appear in supporting documentation. and that
#  any documentation, advertising materials, and other materials related
#  to such distribution and use acknowledge that the software was
#  developed by the University of Southern California, Information
#  Sciences Institute.  The name of the University may not be used to
#  endorse or promote products derived from this software without
#  specific prior written permission.
# 
#  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
#  the suitability of this software for any purpose.  THIS SOFTWARE IS
#  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
#  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# 
#  Other copyrights might apply to parts of this software and are so
#  noted when applicable.
# 
#  Original source contributed by Gaeil Ahn. See below.
#

###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################

#############################################################
#                                                           #
#     File: File for LDP protocol                           #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Jan. 2000      #
#                                                           #
#############################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

Agent/LDP instproc set-mpls-module {module} {
	$self instvar module_
	set module_ $module
}

Agent/LDP instproc get-request-msg {msgid src fec pathvec} {
	$self instvar node_ module_
	
	set pathvec [split $pathvec "_"]
	if {[lsearch $pathvec [$node_ id]] > -1} {
		# loop detected
		set ldpagent [$module_ get-ldp-agent $src]
		$ldpagent new-msgid
		$ldpagent send-notification-msg "LoopDetected" $fec -1
		
		return
	}
	
	set nexthop [$module_ get-nexthop $fec]
	if {$src == $nexthop} {
		#  message from downstream node
		$self request-msg-from-downstream $msgid $src $fec $pathvec
	} else {
		#  message from upstream node
		$self request-msg-from-upstream $msgid $src $fec $pathvec
	}
}

Agent/LDP instproc request-msg-from-downstream {msgid src fec pathvec} {
	$self instvar module_

	#  1. allocate label (upstream label allocation)
	set outlabel [$module_ get-outgoing-label $fec -1]
	if { $outlabel < 0 } {
		if { $fec == $src } {
			# This is a penultimate hop
			set outlabel 0
		} else {
			set outlabel [$module_ new-outgoing-label]
		}     
		$module_ out-label-install $fec -1 $src $outlabel             
	} else {
		set outIface [$module_ get-outgoing-iface $fec -1]
		if { $src != $outIface} {
			# wrong LIB information, let's fix it up
			$module_ out-label-install $fec -1 $src $outlabel
		} 
	}
	
	#  2. reply a allocated label to $src
	set ldpagent [$module_ get-ldp-agent $src]
	$ldpagent new-msgid
	$ldpagent send-mapping-msg $fec $outlabel "*" $msgid

	#  3. trigger label distribution (control-driven-trigger)
	#     do not support on-demand mode in control-driven-trigger
	$module_ ldp-trigger-by-control $fec $pathvec
}

Agent/LDP instproc request-msg-from-upstream {msgid src fec pathvec} {
	$self instvar module_

	#  ldp agent for src
	set ldpagent [$module_ get-ldp-agent $src]

	# This is a egress LSR
	if { [$module_ is-egress-lsr $fec] == 1 } {
                $ldpagent new-msgid
                $ldpagent send-mapping-msg $fec 0 "*" $msgid
                return
	}
	
	set inlabel  [$module_ get-incoming-label $fec -1]
	set outlabel [$module_ get-outgoing-label $fec -1]
	
	# ordered-control mode
	set classifier [$module_ set classifier_]
	if { [$classifier ordered-control?] == 1 } {
		if { $outlabel > -1 } {
			if { $inlabel < 0 } {
				set inlabel [$module_ new-incoming-label]
			}
			$ldpagent new-msgid
			$ldpagent send-mapping-msg $fec inlabel "*" $msgid
		} else {
			$module_ ldp-trigger-by-data $msgid $src $fec $pathvec
		}
		return
	}
	
	#  independent-control mode
	#  1. allocate label (upstream label allocation)
	if { $inlabel < 0 } {
		set inlabel [$module_ new-incoming-label]
		$module_ in-label-install $fec -1 $src $inlabel
	} else {
		set inIface [$module_ get-incoming-iface $fec -1]
		if { $src != $inIface} {
			set classifier [$module_ set classifier_]
			set dontcare [$classifier set dont_care_]
			$module_ in-label-install $fec -1 -1 $dontcare
		} 
	}
	#  2. reply a allocated label to $src
	$ldpagent new-msgid
	$ldpagent send-mapping-msg $fec $inlabel "*" $msgid
	#  3. trigger label distribution (data-driven-trigger)
	#     support on-demand mode in data-driven-trigger
	$module_ ldp-trigger-by-data $msgid $src $fec $pathvec
}

Agent/LDP instproc get-cr-request-msg {msgid src fec pathvec er lspid rc tr pmt} {
	$self instvar node_ module_
	
	#  CR-LDP: used to setup explicit-LSP
	
	set pathvec [split $pathvec "_"]
	if {[lsearch $pathvec [$node_ id]] > -1} {
		# loop detected
		set ldpagent [$module_ get-ldp-agent $src]
		$ldpagent new-msgid
		$ldpagent send-ex-notification-msg "NoRoute" $fec $lspid $tr
		return
	}
	
	# reserve a required resources using $rc
	# if a reservation fails, send notification-msg
	
	#  if a outgoing label for the fec exists ignore this request-msg and
	#  send cr-mapping-msg.
	set outlabel [$module_ get-outgoing-label $lspid]
	
	if { $outlabel > -1 } {
		# This LSP is already established
		set erhop       $fec
		set tunnelLSPid $lspid
		set ret [$module_ admission-control $src $lspid $erhop $tunnelLSPid $tr $pmt]
		if { $ret >= 0 } {
			$module_ send-cr-mapping-after-alloc-label $src $fec $lspid \
					$tunnelLSPid $msgid $pathvec $tr
		} else {
			# ldp agent for src
			set ldpagent [$module_ get-ldp-agent $src]
			$ldpagent new-msgid
			$ldpagent send-ex-notification-msg "RcUnavailable" $fec $lspid $tr
		}
                return
	}
	
	$module_ ldp-trigger-by-explicit-route $msgid $src $fec $pathvec $er \
	    $lspid $rc $tr $pmt
}

Agent/LDP instproc get-cr-mapping-msg {msgid src fec label lspid reqmsgid pathvec tr relabel reiface} {
	$self instvar node_ trace_ldp_ module_
	set pathvec [split $pathvec "_"]
	
	# select a upstream-node to receive a mapping-msg in msgtbl
	# and clear the msgtbl entry
	
	set prvsrc   [$self msgtbl-get-src       $reqmsgid]
	set prvmsgid [$self msgtbl-get-reqmsgid  $reqmsgid]
	set labelop  [$self msgtbl-get-labelop   $reqmsgid]
	
	if {$labelop == 2} {
		set tunnelid [$self msgtbl-get-erlspid $reqmsgid]
	} else {
		set tunnelid -1
	}
	
	set mtr   [$self msgtbl-get-tr   $reqmsgid]
	set hprio [$self msgtbl-get-prio $reqmsgid]
	
	$self msgtbl-clear $reqmsgid
	
	if { $trace_ldp_ } {
		puts "$src -> [$node_ id] : prvsrc($prvsrc)"
		puts "Mapping node([$node_ id]) --> labelop=$labelop $msgid src=$src fec=$fec label=$label lspid=$lspid $reqmsgid tr=$tr"
	}
	
	# allocate outgoing-label
	if { $labelop == 2 } {
		# stack operation
		$module_ out-label-install $fec $lspid $src $label
		$module_ erlsp-stacking $lspid $tunnelid
		if {$hprio != -1} {
			# do not allocate the resource
			set flowid [$module_ new-flowid]
			$module_ erb-append-qosid $lspid $flowid
		}
	} elseif {$labelop == 1} {
		# only pass label
		set ldpagent [$module_ get-ldp-agent $prvsrc]
		$ldpagent new-msgid
		if {$hprio != -1} {
			$ldpagent send-ex-cr-mapping-msg $fec $label $lspid $prvmsgid $pathvec $tr $relabel $reiface
		} else {
			$ldpagent send-cr-mapping-msg    $fec $label $lspid $prvmsgid $pathvec $relabel $reiface
		}
		return
	} else {
		# use the label
		$module_ out-label-install $fec $lspid $src $label
		# allocate the requested resource
		if {$hprio != -1} {
			set flowid [$module_ new-flowid]
			$module_ erb-append-qosid $lspid $flowid
			$module_ resource-reservation $src $mtr $hprio $flowid
		}
	}
	
	# used in reroute-mechanism.
	if { $relabel > -1 } {
		$module_ erb-append-reroute $lspid $relabel $reiface
	}
	if { [$module_ get-reroute-option] == "shortest-dynamic" || [$module_ get-reroute-option] == "simple-dynamic" } {
		set ix [lsearch $pathvec [$node_ id]]
		set len [llength $pathvec]
		set failnext ""
		for {set i [expr $ix + 2]} {$i < $len} {incr i} {
			set failnext [msg-append $failnext [lindex $pathvec $i] "_"]
		}
		if { [string length $failnext] > 0 } {
			$module_ erb-append-failnext $lspid $failnext
		}
	}
	
	if {$prvsrc == [$node_ id]} {
		$module_ give-mapinfo-to-user $lspid
		return
	}
	
	# allocate incoming-label            
	set label [$module_ new-incoming-label]
	$module_ in-label-install $fec $lspid $prvsrc $label

	# find a ldpagent for a upstream-node, continually send a 
	# mapping-msg to the node
	set ldpagent [$module_ get-ldp-agent $prvsrc]
	$ldpagent new-msgid
	if {$hprio != -1} {
                $ldpagent send-ex-cr-mapping-msg $fec $label $lspid $prvmsgid $pathvec $tr $relabel $reiface
	} else {
                $ldpagent send-cr-mapping-msg    $fec $label $lspid $prvmsgid $pathvec $relabel $reiface
	}
}

# msgid    : the msgid of a mapping message sent by dstination
# reqmsgid : the msgid of a request message sent by me before
Agent/LDP instproc get-mapping-msg {msgid src fec label pathvec reqmsgid} {
	$self instvar node_ trace_ldp_ module_
	
	if { $trace_ldp_ } {
		puts "[[Simulator instance] now]: <mapping-msg> $src ->\
                      [$node_ id] : fec($fec), label($label) [$module_ get-nexthop $fec]"
	}
	
	set pathvec [split $pathvec "_"]
	if {[lsearch $pathvec [$node_ id]] > -1} {
		# loop detected
		set ldpagent [$module_ get-ldp-agent $src]
		$ldpagent new-msgid
		$ldpagent send-notification-msg "LoopDetected" $fec -1
		return
	}
	
	set nexthop [$module_ get-nexthop $fec]
	if {$src == $nexthop} {
		#  message from downstream node
		$self mapping-msg-from-downstream $msgid $src $fec $label \
				$pathvec $reqmsgid
	} else {
		#  message from upstream node
		$self mapping-msg-from-upstream $msgid $src $fec $label \
				$pathvec $reqmsgid
	}
}

Agent/LDP instproc mapping-msg-from-downstream {msgid src fec label \
		pathvec reqmsgid} {
	$self instvar node_ module_

	#  allocate label (upstream label allocation)
	$module_ out-label-install $fec -1 $src $label
	if { $reqmsgid > -1 } {
		# downstream-on-demand mode
		
		# 1. find a upstream-node who sent a request-msg in msgtbl
		#    & delete entry in msgtbl
		set prvsrc   [$self msgtbl-get-src      $reqmsgid]
		set prvmsgid [$self msgtbl-get-reqmsgid $reqmsgid]
		$self msgtbl-clear $reqmsgid
		if { $prvsrc == [$node_ id] || $prvsrc < 0} {
			return
		}
		set classifier [$module_ set classifier_]
		if { [$classifier ordered-control?] == 1 } {
			# ordered-control mode

			# 1. allocate incoming-label
			set inlabel [$module_ new-incoming-label]
			$module_ in-label-install $fec -1 $prvsrc $inlabel
			# 2. find a ldpagent for a upstream-node,
			#    send a mapping-msg forward a upstream-node
			set ldpagent [$module_ get-ldp-agent $prvsrc]
			$ldpagent new-msgid
			$ldpagent send-mapping-msg $fec $inlabel -1 $prvmsgid
			return
		}
	} else {
		#  trigger label distribution (control-driven-trigger)
		$module_ ldp-trigger-by-control $fec $pathvec
		return
	}
}

Agent/LDP instproc mapping-msg-from-upstream {msgid src fec label pathvec reqmsgid} {
	$self instvar node_ module_
	
	# check whether or not myself is a next hop of src for a fec.
	set nexthop [$module_ lookup-nexthop $src $fec]
	if { $nexthop != [$node_ id] } {
		return
	}
	
	#  1. allocate label (upstream label allocation)
	set inlabel [$module_ get-incoming-label $fec -1]
	if { $inlabel == -1 } {
		if { [$module_ is-egress-lsr $fec] == 1 } {
			#  message from a penultimate hop
			#  so, send a null label to a penultimate hop
			if { $label != 0 } {
				set ldpagent [$module_ get-ldp-agent $src]
				$ldpagent new-msgid
				$ldpagent send-mapping-msg $fec 0 "*" $msgid
			}
		} else {
			$module_ in-label-install $fec -1 $src $label
		}
	} else {
		set ldpagent [$module_ get-ldp-agent $src]
		if { $reqmsgid < 0 } {
			$ldpagent new-msgid
			$ldpagent send-mapping-msg $fec $inlabel "*" $msgid
		}   
	}    
	#  2. trigger label distribution (data-driven-trigger)
	#     support on-demand mode only in data-driven-trigger
	if { $reqmsgid < 0 } {
		$module_ ldp-trigger-by-data -1 $src $fec $pathvec
	} else {   
		# if ($reqmsgid > -1) then delete a entry in msgid table
		$self msgtbl-clear $reqmsgid
	}    
}

Agent/LDP instproc get-notification-msg {src status fec lspid tr} {
	$self instvar node_ trace_ldp_ module_

	# Notification-msg (in case of Loop Detection, No Route)
	if { $trace_ldp_ } {
		puts "Notification($src->[$node_ id]): $status src=$src lspid=$lspid"
	}
	set msgid [$self msgtbl-get-msgid -1 $lspid -1]
	if {$msgid > -1} {
		set prvsrc [$self msgtbl-get-src  $msgid]
		set mtr    [$self msgtbl-get-tr   $msgid]
		set hprio  [$self msgtbl-get-prio $msgid]
		$self msgtbl-clear $msgid
		if { $hprio != -1} {
			$module_ fail-admission-control $src $mtr
		}
		if { $prvsrc < -1 || $prvsrc == [$node_ id] } {
			return
		}
		set ldpagent [$module_ get-ldp-agent $prvsrc]
		$ldpagent new-msgid
		if { $hprio != -1} {
			$ldpagent send-ex-notification-msg $status $fec $lspid $tr
		} else {
			$ldpagent send-notification-msg    $status $fec $lspid
		}
	} elseif {$lspid > -1} {
		set iniface [$module_ get-incoming-iface $lspid]
		if { $iniface > -1} {
			set ldpagent [$module_ get-ldp-agent $iniface]
			$ldpagent new-msgid
			$ldpagent send-ex-notification-msg $status $fec $lspid $tr
		} else {
			$module_ give-notinfo-to-user $status $lspid $tr
			return
		}
	}
}

Agent/LDP instproc get-withdraw-msg {src fec lspid tr} {
	$self instvar module_
	
	# withdraw msg : downstream -> upstream
	set lspid_list -1
	if { $lspid > -1 } {
		set lspid_list [$module_ get-spliced-lsp $lspid]
	}
	if {$lspid > -1} {
		set outiface  [$module_ get-outgoing-iface $lspid]
	} else {
		set outiface  [$module_ get-outgoing-iface $fec -1]
	}
	
	$module_ out-label-clear $fec $lspid
	
	if {$src == $outiface} {  
		if {$lspid > -1} {
			set flowid [$module_ get-flowid-for-lspid $lspid]
			if { $flowid > 0} {
				$module_ resource-release $src $flowid
			}
		}
		#checking the spliced LSP
		if { $lspid > -1 } {
			if { $lspid_list != "-1" } {
				set lspid_list [split $lspid_list "_"]
				set lspid_listlen [llength $lspid_list]
				
				for {set i 0} {$i < $lspid_listlen} {incr i 1} {
					set s_lspid [lindex $lspid_list $i]
					
					set s_fec [$module_ get-fec-for-lspid $s_lspid]
					set s_iface [$module_ get-incoming-iface $s_lspid]
					$module_ in-label-clear $s_fec $s_lspid
					
					set ldpagent [$module_ get-ldp-agent $s_iface]
					$ldpagent new-msgid
					$ldpagent send-withdraw-msg $s_fec $s_lspid
				}
			}
		}
		
		if {$lspid > -1} {
			set inlabel [$module_ get-incoming-label $lspid]
		} else {
			set inlabel [$module_ get-incoming-label $fec -1]
		}
		
		if {$inlabel > 0} {
			$module_ ldp-trigger-by-ex-withdraw $fec $lspid $src
		} else {
			if { $lspid > -1 } {
				$module_ give-widinfo-to-user $lspid
			}
		}
	} else {
		# ignore this message
	}	
}

Agent/LDP instproc get-release-msg {src fec lspid tnl_lspid} {
	$self instvar module_
	
	# release msg : upstream --> downstream
	if {$lspid > -1} {
		set iniface  [$module_ get-incoming-iface $lspid]
		set inlabel  [$module_ get-incoming-label $lspid]
		set outlabel [$module_ get-outgoing-label $lspid]
	} else {
		set iniface  [$module_ get-incoming-iface $fec -1]
		set inlabel  [$module_ get-incoming-label $fec -1]
		set outlabel [$module_ get-outgoing-label $fec -1]
	}
	
	if { $inlabel > -1 } {
		if {$iniface == $src} {
			$module_ in-label-clear $fec $lspid 
			$module_ ldp-trigger-by-release $fec $lspid
		}
	} else {
		if { $tnl_lspid > -1 } {	
			set oiface [$module_ get-outgoing-iface $tnl_lspid]
			if { $oiface > -1 } {
				set ldpagent [$module_ get-ldp-agent $oiface]
				$ldpagent new-msgid
				$ldpagent send-release-msg $fec $lspid $tnl_lspid
				return
			}
		}
		if { $lspid > -1 } {
			$module_ give-widinfo-to-user $lspid
		}
	}
}

Agent/LDP instproc trace-ldp-packet {src_addr msgtype msgid fec \
				     label pathvec lspid er rc \
				     reqmsgid status atime} {
	$self instvar node_
	puts "$atime [$node_ id]: $src_addr ($msgtype $msgid) $fec $label $pathvec  \[$reqmsgid $status\]  \[$lspid $er $rc\]"
}

Agent/LDP instproc send-notification-msg {status fec lspid} {
	$self set fid_ 991
	$self notification-msg $status $fec $lspid "*"
}

Agent/LDP instproc send-ex-notification-msg {status fec lspid tr} {
	$self set fid_ 991
	$self notification-msg $status $fec $lspid $tr
}

Agent/LDP instproc send-request-msg {fec pathvec} {
	$self set fid_ 992
	$self request-msg $fec $pathvec
}

Agent/LDP instproc send-mapping-msg {fec label pathvec reqmsgid} {
	$self set fid_ 993
	set ns [Simulator instance]
	$self mapping-msg $fec $label $pathvec $reqmsgid
}

Agent/LDP instproc send-withdraw-msg {fec lspid} {
	$self set fid_ 994
	$self withdraw-msg $fec $lspid
}

Agent/LDP instproc send-release-msg args {
	$self set fid_ 995
	set nargs [llength $args]
	set fec   [lindex $args 0]
	set lspid [lindex $args 1]
	if { $nargs == 3 } {
		set er [lindex $args 2]
		$self release-msg $fec $lspid $er
	} else {
                $self release-msg $fec $lspid  
	}
}

Agent/LDP instproc send-cr-request-msg {fec pathvec er lspid rc} {
	$self set fid_ 992
	$self cr-request-msg $fec $pathvec $er $lspid $rc
}

Agent/LDP instproc send-ex-cr-request-msg {fec pathvec er lspid rc tr pmt} {
	$self set fid_ 992
	$self cr-request-msg $fec $pathvec $er $lspid $rc $tr $pmt
}

Agent/LDP instproc send-cr-mapping-msg args {
	set nargs    [llength $args]	
	set fec      [lindex $args 0]
	set label    [lindex $args 1]
	set lspid    [lindex $args 2]
	set prvmsgid [lindex $args 3]
	set pathvec  [lindex $args 4]
	set tr       "*"
	set relabel -1
	set reiface -1
	if { $nargs == 7 } {
		set relabel [lindex $args 5]
		set reiface [lindex $args 6]
	}
	$self send-ex-cr-mapping-msg $fec $label $lspid $prvmsgid \
			$pathvec "*" $relabel $reiface
}

Agent/LDP instproc send-ex-cr-mapping-msg args {
        $self instvar module_ node_
	$self set fid_ 993	
	set nargs    [llength $args]
	set fec      [lindex $args 0]
	set inlabel  [lindex $args 1]
	set lspid    [lindex $args 2]
	set prvmsgid [lindex $args 3]
	set pathvec  [lindex $args 4]
	set tr       [lindex $args 5]
	set relabel -1
	if { $nargs == 8 } {
		set relabel [lindex $args 6]
		set reiface [lindex $args 7]
	}
	if { [$module_ is-reroute-egress-lsr] == 1 } {
		set reiface [$node_ id]
		set relabel $inlabel
		$self cr-mapping-msg $fec $inlabel $lspid $prvmsgid \
				$pathvec $tr $relabel $reiface
	} elseif { $relabel > -1} {
		$self cr-mapping-msg $fec $inlabel $lspid $prvmsgid \
				$pathvec $tr $relabel $reiface
		
	} else {
		$self cr-mapping-msg $fec $inlabel $lspid $prvmsgid \
				$pathvec $tr
	}
}

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
