# -*-	Mode:tcl; tcl-indent-level:8; tab-width:8; indent-tabs-mode:t -*-
#
#  Time-stamp: <2000-09-11 15:10:21 haoboy>
# 
#  Copyright (c) 2000 by the University of Southern California
#  All rights reserved.
# 
#  Permission to use, copy, modify, and distribute this software and its
#  documentation in source and binary forms for non-commercial purposes
#  and without fee is hereby granted, provided that the above copyright
#  notice appear in all copies and that both the copyright notice and
#  this permission notice appear in supporting documentation. and that
#  any documentation, advertising materials, and other materials related
#  to such distribution and use acknowledge that the software was
#  developed by the University of Southern California, Information
#  Sciences Institute.  The name of the University may not be used to
#  endorse or promote products derived from this software without
#  specific prior written permission.
# 
#  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
#  the suitability of this software for any purpose.  THIS SOFTWARE IS
#  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
#  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# 
#  Other copyrights might apply to parts of this software and are so
#  noted when applicable.
# 
#  Original source contributed by Gaeil Ahn. See below.
#

###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################

#
# XXX MPLS is NOT compatible with hierarchical routing, because:
#
# 1) MPLS classifier is explicitly coupled with unicast address classifier,
#    which does not support hierarchical routing. The reason is MPLS needs 
#    to check whether a route update is an update, a new entry, or a 
#    "no change". This is not as straightforward and efficient when the 
#    MPLS classifier is decoupled from the unicast address classifier, because
#    through add-route{} interface, Node can only tell this module that a new
#    route is added, but not whether this is an update, a new entry, or 
#    a "no change". I don't have a clean solution yet. - haoboy
#

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

RtModule/MPLS instproc register { node } {
	$self instvar classifier_
	$self instvar failure_type_
	
	$self attach-node $node
	$node route-notify $self
	$node port-notify $self
        
	set classifier_ [new Classifier/Addr/MPLS]
        $classifier_ set-node $node $self
	$node install-entry $self $classifier_ 0
	$self attach-classifier $classifier_
	
	set failure_type_ 0
}

RtModule/MPLS instproc enable-data-driven {} {
	[$self set classifier_] enable-data-driven
}

RtModule/MPLS instproc enable-control-driven {} {
	[$self set classifier_] enable-control-driven
}

RtModule/MPLS instproc enable-on-demand {} {
	[$self set classifier_] enable-on-demand
}

RtModule/MPLS instproc enable-ordered-control {} {     
	[$self set classifier_] enable-ordered-control
}

RtModule/MPLS instproc make-ldp {} {
	set ldp [new Agent/LDP]
	$self attach-ldp $ldp
	$ldp set-mpls-module $self
	[$self node] attach $ldp
	return $ldp
}

RtModule/MPLS instproc exist-fec {fec phb} {
        return [[$self set classifier_] exist-fec $fec $phb]
}

RtModule/MPLS instproc get-incoming-iface args {
        set nargs [llength $args]
        if { $nargs == 1 } {
             set lspid  [lindex $args 0]
             return [[$self set classifier_] GetInIface -1 -1 $lspid]
        } elseif { $nargs == 2 } {
             set fec    [lindex $args 0]
             set phb    [lindex $args 1]
             return [[$self set classifier_] GetInIface $fec $phb -1]
        } else {
		puts "file: ns-mpls-node.tcl"
		puts "RtModule/MPLS instproc get-incoming-iface"
		puts "error: wrong argument number"
		puts $args
		exit
        }
}

RtModule/MPLS instproc get-incoming-label args {
        set nargs [llength $args]
        if { $nargs == 1 } {
		set lspid  [lindex $args 0]
		return [[$self set classifier_] GetInLabel -1 -1 $lspid]
        } elseif { $nargs == 2 } {
		set fec    [lindex $args 0]
		set phb    [lindex $args 1]
		return [[$self set classifier_] GetInLabel $fec $phb -1]
        } else {
		puts "file: ns-mpls-node.tcl"
		puts "RtModule/MPLS instproc get-incoming-label"
		puts "error: wrong argument number"
		puts $args
		exit
	}
}

RtModule/MPLS instproc get-outgoing-label args {
        set nargs [llength $args]
        if { $nargs == 1 } {
             set lspid  [lindex $args 0]
             return [[$self set classifier_] GetOutLabel -1 -1 $lspid]
        } elseif { $nargs == 2 } {
             set fec    [lindex $args 0]
             set phb    [lindex $args 1]
             return [[$self set classifier_] GetOutLabel $fec $phb -1]
        } else {
		puts "file: ns-mpls-node.tcl"
		puts "RtModule/MPLS instproc get-outgoing-label"
		puts "error: wrong argument number"
		puts $args
		exit
        }
}

RtModule/MPLS instproc get-outgoing-iface args {
	set nargs [llength $args]
        if { $nargs == 1 } {
		set lspid  [lindex $args 0]
		return [[$self set classifier_] GetOutIface -1 -1 $lspid]
        } elseif { $nargs == 2 } {
		set fec [lindex $args 0]
		set phb [lindex $args 1]
		return [[$self set classifier_] GetOutIface $fec $phb -1]
        } else {
		puts "file: ns-mpls-node.tcl"
		puts "RtModule/MPLS instproc get-outgoing-iface"
		puts "error: wrong argument number"
		puts "$args"
		exit
        }	
}

RtModule/MPLS instproc get-fec-for-lspid {lspid} {
        return [[$self set classifier_] get-fec-for-lspid $lspid]
}

RtModule/MPLS instproc get-tunnelid-for-lspid {lspid} {
        return [[$self set classifier_] get-tunnelid-for-lspid $lspid]
}

RtModule/MPLS instproc in-label-install {fec lspid iface label} {
	set dontcare [Classifier/Addr/MPLS dont-care]
	$self label-install $fec $lspid $iface $label $dontcare $dontcare
}

RtModule/MPLS instproc out-label-install {fec lspid iface label} {
	set dontcare [Classifier/Addr/MPLS dont-care]
	$self label-install $fec $lspid $dontcare $dontcare $iface $label
}

RtModule/MPLS instproc in-label-clear {fec lspid} {
	set dontcare [Classifier/Addr/MPLS dont-care]
	$self label-clear $fec $lspid -1 -1 $dontcare $dontcare
}

RtModule/MPLS instproc out-label-clear {fec lspid} {
	set dontcare [Classifier/Addr/MPLS dont-care]
	$self label-clear $fec $lspid $dontcare $dontcare -1 -1
}

RtModule/MPLS instproc label-install {fec lspid iif ilbl oif olbl} {
        [$self set classifier_] LSPsetup $fec $lspid $iif $ilbl $oif $olbl
}

RtModule/MPLS instproc label-clear {fec lspid iif ilbl oif olbl} {
        [$self set classifier_] LSPrelease $fec $lspid $iif $ilbl $oif $olbl
}

RtModule/MPLS instproc splice-lsp {fistlspid lastlspid} {
	[$self set classifier_] splice-lsp $fistlspid $lastlspid
}

RtModule/MPLS instproc get-spliced-lsp {lspid} {
	set spliced_lsp_list [[$self set classifier_] get-spliced-lsp $lspid]
	return $spliced_lsp_list
}

RtModule/MPLS instproc bind-flow-erlsp {fec phb lspid} {
        set ret [[$self set classifier_] ErLspBinding $fec $phb $lspid]
}

RtModule/MPLS instproc flow-erlsp-install {fec phb lspid} {
        [$self set classifier_] ErLspBinding $fec $phb $lspid
}

RtModule/MPLS instproc erlsp-stacking {erlspid tunnelid} {
        [$self set classifier_] ErLspStacking -1 $erlspid -1 $tunnelid
}

RtModule/MPLS instproc aggregate-flows {fineFec coarseFec} {
        [$self set classifier_] FlowAggregation $fineFec -1 $coarseFec -1
}

RtModule/MPLS instproc flow-aggregation {fineFec finePhb coarseFec coarsePhb} {
        [$self set classifier_] FlowAggregation $fineFec \
			$finePhb $coarseFec $coarsePhb
}

RtModule/MPLS instproc reroute-binding {fec phb relspid} {
        [$self set classifier_] aPathBinding $fec $phb -1 $relspid
}

RtModule/MPLS instproc reroute-fec-binding {fec phb relspid} {
        [$self set classifier_] aPathBinding $fec $phb -1 $relspid
}

RtModule/MPLS instproc reroute-lsp-binding {lspid relspid} {
        [$self set classifier_] aPathBinding $lspid $relspid
}

RtModule/MPLS instproc secondary-fec-binding {fec phb relspid} {
	[$self set classifier_] sec-lsp-binding $fec $phb -1 $relspid
}

RtModule/MPLS instproc secondary-lsp-binding {lspid relspid} {
	set ret [[$self set classifier_] sec-lsp-binding $lspid $relspid]
}

RtModule/MPLS instproc erb-append-reroute {lspid relabel reiface} {
	[$self set classifier_] erb-append-reroute $lspid $relabel $reiface
}

RtModule/MPLS instproc erb-append-failnext {lspid failnext} {
	[$self set classifier_] erb-append-failnext $lspid $failnext
}

Node instproc get-rtmetric {fec} {
        set routeobject [$self rtObject?]
        return [$routeobject metric? $fec]
}

RtModule/MPLS instproc lookup-nexthop {node fec} {
        set ns [Simulator instance]
        set routingtable [$ns get-routelogic]
        set nexthop [$routingtable lookup $node $fec]
        return $nexthop
}

RtModule/MPLS instproc get-nexthop {fec} {
	# find a next hop for fec
        set nodeid [[$self node] id]
        set nexthop [$self lookup-nexthop $nodeid $fec]
        return $nexthop
}

RtModule/MPLS instproc get-mpls-neighbor {} {
	set ns [Simulator instance]
	set alllsr [$ns set linked_mplsnodes_]
	set lsr_list ""
	set len [llength [[$self node] set neighbor_]]
	for {set i 0} {$i < $len} {incr i} {
		set tnode [lindex [[$self node] set neighbor_] $i]
		if { [lsearch $alllsr $tnode] > -1 } {
			lappend lsr_list $tnode
		}
	}
	return $lsr_list
}

RtModule/MPLS instproc get-link-status {hop} {
	if {$hop < 0} {
		return "down"
	}
	set nexthop [$self get-nexthop $hop]
	if {$nexthop == $hop} {
		set status "up"
	} else {
		set status "down"
	}
	return $status
}

RtModule/MPLS instproc is-egress-lsr {fec} {
        # Determine whether I am a egress-lsr for fec. 
        if { [[$self node] id] == $fec } {
		return  "1"
        }
        set nexthopid [$self get-nexthop $fec]
        if { $nexthopid < 0 } {
		return "-1"
        }
        set nexthop [[Simulator instance] get-node-by-id $nexthopid]
	if { [$nexthop get-module "MPLS"] == "" } {
		return  "1"
        } else {
		return  "-1"
        }
}

# distribute labels based on routing protocol
RtModule/MPLS instproc ldp-trigger-by-routing-table {} {
        if { [[$self set classifier_] control-driven?] != 1 } {
		return
        }
        set ns [Simulator instance]
	for {set i 0} {$i < [$ns get-number-of-nodes]} {incr i} {
		# select a node
		set host [$ns get-node-by-id $i]
		if { [$self is-egress-lsr [$host id]] == 1 } {
			# distribute label
			set nlsr       [lindex [$self get-mpls-neighbor] 0]
			set nlsrModule [eval $nlsr get-module "MPLS"]
			if { [$nlsrModule get-nexthop [$host id]] >= 0 } {
				$self ldp-trigger-by-control [$host id] *
			} else {
				# routing information is not stable
				$self register-ldp-trigger [$host id] $nlsrModule
			}
		}
        }
}

RtModule/MPLS instproc register-ldp-trigger {fec nlsr} {
        if { [$nlsr get-outgoing-iface $fec -1] != [$self id] } {
		$self ldp-trigger-by-control $fec *
		
		set link_ptr [$self get-link-ptr $nlsr]
		set delay [expr [$link_ptr get-link-delay] * 2]
		set ns [Simulator instance]
		$ns at [expr [$ns now]+$delay] "$self register-ldp-trigger $fec $nlsr"
        }
}

RtModule/MPLS instproc ldp-trigger-by-control {fec pathvec} {
	lappend pathvec [[$self node] id]
	set inlabel [$self get-incoming-label $fec -1]
	set nexthop [$self get-nexthop $fec]
	set ldpagents [lsort [$self get-ldp-agents]]
	for {set i 0} {$i < [llength $ldpagents]} {incr i 1} {
		set ldpagent [lindex $ldpagents $i]
		if { [$ldpagent peer-ldpnode] == $nexthop } {
			continue
		}
		if { $inlabel == -1 } {
			if { [$self is-egress-lsr $fec] == 1 } {
				# This is egress LSR
				set inlabel 0
			} else {
				set inlabel [$self new-incoming-label]
				$self in-label-install $fec -1 -1 $inlabel
			}
		}
		$ldpagent new-msgid
		$ldpagent send-mapping-msg $fec $inlabel $pathvec -1
	}
}

RtModule/MPLS instproc ldp-trigger-by-data {reqmsgid src fec pathvec} {
	if { [$self is-egress-lsr $fec] == 1 } {
		# This is a egress node
		return
	}
	set outlabel [$self get-outgoing-label $fec -1]
	if { $outlabel > -1  } {
		# reroute check !! to establish alternative path
		set outiface [$self get-outgoing-iface $fec -1]
		if { [$self get-link-status $outiface] == "up" } {
			# LSP was already setup
			return
		}
	}
	lappend pathvec [[$self node] id]      
	set nexthop [$self get-nexthop $fec]

	set ldpagents [lsort [$self get-ldp-agents]]
	
	for {set i 0} {$i < [llength $ldpagents]} {incr i 1} {
		set ldpagent [lindex $ldpagents $i]
		
		if { [$ldpagent peer-ldpnode] == $nexthop } {
			if {$reqmsgid > -1} {
				# on-demand mode
				set msgid [$ldpagent msgtbl-get-msgid $fec -1 -1]
				if { $msgid < 0 } {
					# have to make new request-msg
					set newmsgid [$ldpagent new-msgid]
					$ldpagent msgtbl-install $newmsgid $fec -1 \
							$src $reqmsgid
					$ldpagent send-request-msg $fec $pathvec
				} else {
					# $self is already trying to setup a LSP
					# So, do not need to send a request-msg
					# but check whether src information is correct.
					if { $src == [[$self node] id] } {
						return
					}
					set prvsrc [$ldpagent msgtbl-get-src  $msgid]
					if { $prvsrc != $src }  {
						$ldpagent msgtbl-clear $msgid
						$ldpagent msgtbl-install $msgid $fec -1 $src $reqmsgid
					}
				}
			} else {
				# not on-demand mode
				if {$fec == $nexthop} {
					# This is a penultimate hop
					set outlabel 0
				} else {
					set outlabel [$self new-outgoing-label]
				}
				$self out-label-install $fec -1 $nexthop $outlabel
				$ldpagent new-msgid
				$ldpagent send-mapping-msg $fec $outlabel $pathvec -1
			}
			return
		}
	}
}

RtModule/MPLS instproc setup-erlsp {fec er lspid} {
      $self make-explicit-route $fec $er $lspid -1
}

RtModule/MPLS instproc setup-crlsp {fec er lspid TRate BSize PSize SPrio HPrio} {
	# Traffic_Parameters: Frequency(0), Weight(1), PDR(2), PBS(3), CDR(4), CBS(5), EBS(6) 
	
	set tr "*_*_*_*"
	set CDR   [parse-bw $TRate]
	set BSize [parse-bw $BSize]
	set PSize [parse-bw $PSize]
	set CBS   [expr $BSize/$PSize]
	set tr [msg-append $tr $CDR "_"]
	set tr [msg-append $tr $CBS "_"]
	set tr [msg-append $tr "*"  "_"]
	
	if {$CDR < 0} {
		puts "Error : specify TokenRate"
		exit
	} elseif {$CBS < 0} {
		puts "Error : specify BucketSize and PacketSize"
		exit
	}
	
	set pmt $SPrio
	set pmt [msg-append $pmt $HPrio "_"]

	if {$SPrio > 7} {
		puts "Error : SetupPriority is a value between 0 and 7"
		exit
	} elseif {$HPrio > 7} { 
		puts "Error : HoldPriority is a value between 0 and 7"
		exit
	}
	
	$self ldp-trigger-by-explicit-route -1 [[$self node] id] $fec "*" $er $lspid -1 $tr $pmt 
}

RtModule/MPLS instproc make-explicit-route {fec er lspid rc} {
	$self ldp-trigger-by-explicit-route -1 [[$self node] id] $fec "*" $er $lspid $rc "*" "*"
}

RtModule/MPLS instproc ldp-trigger-by-explicit-route {reqmsgid src fec pathvec er lspid rc tr pmt} {
	$self instvar classifier_
	
	set outlabel [$self get-outgoing-label $lspid]
	
	if { $outlabel > -1 } {
		# LSP was already setup
		return
	}
	if { [[$self node] id] != $src && [[$self node] id] == $fec } {
		# This is a egress node
		set tunnelLSPid -1
		$self send-cr-mapping-after-alloc-label $src $fec $lspid $tunnelLSPid $reqmsgid $pathvec $tr
		return
	}
	
	lappend pathvec [[$self node] id]
	set er [split $er "_"]
	set erlen [llength $er]
	for {set i 0} {$i <= $erlen} {incr i 1} {
		if { $i != $erlen } {
			set erhop [lindex $er $i]
		} else {
			set erhop $fec
		}
		
		if { [string match "L*" $erhop] == 1 } {
			# This is the ER-Hop and LSPid (ER-LSP)
			set erhop [string trimleft $erhop L]
			set lspidFEC [$self get-fec-for-lspid $erhop]
			
			if { $src == [[$self node] id] } {
				continue
			} elseif { $lspidFEC == -1 } {
				continue
			} elseif { $lspidFEC == $fec } {
				# splice two LSPs
				set inlabel  [$self get-incoming-label $erhop]
				set outlabel [$self get-outgoing-label $erhop]
				
				if { $outlabel > -1 || $inlabel > -1} {
					set tunnelLSPid $erhop
					set ret [$self admission-control $src $lspid $erhop $tunnelLSPid $tr $pmt]
					if { $ret >= 0 } {
						$self send-cr-mapping-after-alloc-label $src $fec $lspid \
						    $tunnelLSPid $reqmsgid $pathvec $tr
					}
					return
				} else {
					puts "error : no Tunnel(lspid=$erhop) for the establishment of ER-LSP(lspid=$lspid)"
					continue
				}
			} else {
				# do stack operation
				set outlabel [$self get-outgoing-label $erhop]
				set existExplicitPeer [$self exist-ldp-agent $lspidFEC]
				
				if { $outlabel > -1 && $existExplicitPeer == "1" } {
					set tunnelLSPid $erhop
					set erhop       $lspidFEC
					set ok [$self admission-control $src $lspid $erhop $tunnelLSPid $tr $pmt]
					if { $ok >= 0 } {
						$self send-reqmsg-to-nexter $reqmsgid $src $fec $pathvec $er \
						    $lspid $rc $erhop $tunnelLSPid $tr $pmt
					}
				} elseif { $outlabel > -1 && $existExplicitPeer == "0" } {
					set tunnelLSPid $erhop
					set ok [$self admission-control $src $lspid $erhop $tunnelLSPid $tr $pmt]
					if { $ok >= 0 } {
						 $self send-reqmsg-through-tunnel $reqmsgid $src $fec $pathvec $er \
						     $lspid $rc $tunnelLSPid $tr $pmt
					}
					return
				} else {
					# This erhop is already processed, So, let's find the next erhop
					continue
				}
			}
		} else {
			if { [lsearch $pathvec $erhop] < 0 } {
				if { [$classifier_ set enable_reroute_] == 1 } {
					set len [llength [[$self node] set neighbor_]]
					for {set i 0} {$i < $len} {incr i} {
						set nnode [lindex [[$self node] set neighbor_] $i]
						if { [$nnode id] == $erhop } {
							break
						}
					}
					if { $i != $len } {
						if { [$self get-link-status $erhop] == "down" } {
							set ldpagent [$self get-ldp-agent $src]
							$ldpagent new-msgid
							$ldpagent send-ex-notification-msg "NoRoute" $fec $lspid $tr
							return
						}
					}
				}
				set tunnelLSPid -1
				set ok [$self admission-control $src $lspid $erhop $tunnelLSPid $tr $pmt]
				if { $ok >= 0 } {
					$self send-reqmsg-to-nexter $reqmsgid $src $fec $pathvec $er $lspid $rc \
					    $erhop $tunnelLSPid $tr $pmt
				}
				return
			}
		}
	}
	
	set ldpagent [$self get-ldp-agent $src]
	if { $ldpagent != "" } {
		$ldpagent new-msgid
		$ldpagent send-ex-notification-msg "NoRoute" $fec $lspid $tr
	}
}

RtModule/MPLS instproc send-reqmsg-through-tunnel {reqmsgid src fec pathvec er lspid rc tunnelid tr pmt} {
	set nexthop [$self get-outgoing-iface $tunnelid]
	set iiface  [$self get-incoming-iface $tunnelid]
	set ldpagent [$self get-ldp-agent $nexthop]
	set working [$ldpagent msgtbl-get-msgid $fec $lspid $src]
	
	if { $working < 0 } {
		# have to make new request-msg
		set newmsgid [$ldpagent new-msgid]
		$ldpagent msgtbl-install $newmsgid $fec $lspid $src $reqmsgid
		
		# QoS
		set hprio [$self pmtparms-to-hprio $pmt]
		if {$hprio > -1} {
			$ldpagent msgtbl-append-qos $newmsgid $tr $hprio
		}
		
		# determine whether labelpass or labelstack
		set inlabel  [$self get-incoming-label $tunnelid]
		
		if { $iiface == $src && $inlabel > -1 } {
			$ldpagent msgtbl-set-labelpass $newmsgid
		} else {
			$ldpagent msgtbl-set-labelstack $newmsgid $tunnelid
		}
		
		if {$hprio > -1} {
			$ldpagent send-ex-cr-request-msg $fec $pathvec $er $lspid $rc $tr $pmt
		} else {
			$ldpagent send-cr-request-msg    $fec $pathvec $er $lspid $rc
		}
		
	} else {
		# $self is already trying to setup a LSP
		# So, do need not to send a request-msg
	}
}

RtModule/MPLS instproc send-reqmsg-to-nexter {reqmsgid src fec pathvec er lspid rc erhop tunnelid tr pmt} {
	set nexthop [$self get-nexthop $erhop]
	
	if { [$self is-egress-lsr $nexthop] == 1 } {
		# MPLS-node that can process the reqested er-LSP does not exist
		if { $erhop == $fec } {
			# This is a final node
			$self send-cr-mapping-after-alloc-label $src $fec $lspid $tunnelid $reqmsgid $pathvec $tr
			return
		} else {
			# wrong er list
			set ldpagent [$self get-ldp-agent $src]
			$ldpagent new-msgid
			$ldpagent send-ex-notification-msg "NoRoute" $fec $lspid $tr
			return
		}
	} else {
		set ldpagent [$self get-ldp-agent $nexthop]
		set working [$ldpagent msgtbl-get-msgid $fec $lspid $src]
		if { $working < 0 } {
			# have to make new request-msg
			set newmsgid [$ldpagent new-msgid]
			set id [$ldpagent msgtbl-install $newmsgid $fec $lspid $src $reqmsgid]
			
			# QoS
			set hprio [$self pmtparms-to-hprio $pmt]
			if {$hprio > -1} {
				$ldpagent msgtbl-append-qos $newmsgid $tr $hprio
			}
			
			# determine whether labelpass or labelstack
			if { $tunnelid > -1 } {
				set olabel  [$self get-outgoing-label $tunnelid]
				if { $olabel == 0 } {
					$ldpagent msgtbl-set-labelpass $newmsgid
				} else {
					$ldpagent msgtbl-set-labelstack $newmsgid $tunnelid
				}
			}
			
			if {$hprio > -1} {
				$ldpagent send-ex-cr-request-msg $fec $pathvec $er $lspid $rc $tr $pmt
			} else {
				$ldpagent send-cr-request-msg    $fec $pathvec $er $lspid $rc
			}
		} else {
			# $self is already trying to setup a LSP
			# So, need not to send a request-msg
		}
	} 
}

RtModule/MPLS instproc send-cr-mapping-after-alloc-label {src fec lspid tunnelid msgid pathvec tr} {
	# if (flowid=0), No request for QoS
	set ldpagent [$self get-ldp-agent $src]
	$ldpagent new-msgid
	
	# allocate a incoming Label
	if { $tunnelid > 0 } {
		set inlabel  [$self get-incoming-label $tunnelid]
		set outlabel [$self get-outgoing-label $tunnelid]
		
		if { $outlabel > -1 || $inlabel > -1} {
			$self splice-lsp $lspid $tunnelid
			set inlabel [$self new-incoming-label]
			$self in-label-install -1 $lspid $src $inlabel
		} else {
			puts "Error : wrong Tunnel ID"
			exit
		}
	} else {
		set inlabel  [$self get-incoming-label $lspid]
		set outlabel [$self get-outgoing-label $lspid]
		
		if { $outlabel > -1 || $inlabel > -1} {
			puts "Error : duplicated LSPID"
			exit
		} else {
			# This is Egress LSR
			set inlabel 0
		}
	}
	
	if { [$self trparms-to-bw $tr] == -1 } {
		$ldpagent send-cr-mapping-msg $fec $inlabel $lspid $msgid $pathvec
	} else {
		$ldpagent send-ex-cr-mapping-msg $fec $inlabel $lspid $msgid $pathvec $tr
	}
}

RtModule/MPLS instproc send-ldp-withdraw-msg {fec} {
      $self ldp-trigger-by-withdraw $fec -1
}

RtModule/MPLS instproc send-crldp-withdraw-msg {lspid} {
      set fec [$self get-fec-for-lspid $lspid]
      $self ldp-trigger-by-withdraw $fec $lspid
}

RtModule/MPLS instproc ldp-trigger-by-withdraw {fec lspid} {
      if {$lspid > -1} {
           set outiface [$self get-outgoing-iface $lspid]
      } else {
           set outiface [$self get-outgoing-iface $fec -1]
      }
      $self ldp-trigger-by-ex-withdraw $fec $lspid $outiface
}

RtModule/MPLS instproc ldp-trigger-by-ex-withdraw {fec lspid nexthop} {
	if {$lspid > -1} {
		set iniface [$self get-incoming-iface $lspid]
	} else {
		set iniface [$self get-incoming-iface $fec -1]
	}
	
	$self in-label-clear $fec $lspid
	
	if {$iniface > -1} {
		set ldpagent [$self get-ldp-agent $iniface]
		if { $ldpagent != "" } {
			$ldpagent new-msgid
			$ldpagent send-withdraw-msg $fec $lspid
		}
	} else {
		# several upstream nodes may share a label.
		# so inform all upstream nodes to withdraw the label
		if {$nexthop < 0} {
			set nexthop [$self get-nexthop $fec]
		}
		set ldpagents [lsort [$self get-ldp-agents]]
		for {set i 0} {$i < [llength $ldpagents]} {incr i 1} {
			set ldpagent [lindex $ldpagents $i]
			if { [$ldpagent peer-ldpnode] == $nexthop } {
				continue
			}
			$ldpagent new-msgid
			$ldpagent send-withdraw-msg $fec $lspid
		}
	}   
}

RtModule/MPLS instproc send-ldp-release-msg {fec} {
	$self ldp-trigger-by-release $fec -1
}

RtModule/MPLS instproc send-crldp-release-msg {lspid} {
	set fec [$self get-fec-for-lspid $lspid]
	$self ldp-trigger-by-release $fec $lspid
}

RtModule/MPLS instproc ldp-trigger-by-release {fec lspid} {
	if {$lspid > -1} {
		set outlabel [$self get-outgoing-label $lspid]
	} else {
		set outlabel [$self get-outgoing-label $fec -1]
	}
	
	if {$outlabel < 0} {
		# splicing LSP
		return
	}
	
	if {$lspid > -1} {
		set nexthop [$self get-outgoing-iface $lspid]
	} else {
		set nexthop [$self get-outgoing-iface $fec -1]
	}      
	
	# check if there is the allocated resource for the LSP($lspid)
	if { $lspid > 0 } {
                set flowid  [$self get-flowid-for-lspid $lspid]
                if { $flowid > 0} {
			$self resource-release $nexthop $flowid
                }
	}
	
	if { $lspid > -1 } {
		set lspid_list [$self get-spliced-lsp $lspid]
		if { $lspid_list != "-1" } {
			set lspid_list [split $lspid_list "_"]
			set lspid_listlen [llength $lspid_list]
			
			for {set i 0} {$i < $lspid_listlen} {incr i 1} {
				set s_lspid [lindex $lspid_list $i]
				set s_fec [$self get-fec-for-lspid $s_lspid]
				set s_iface  [$self get-incoming-iface $s_lspid]
				$self in-label-clear $s_fec $s_lspid
				set ldpagent [$self get-ldp-agent $s_iface]
				$ldpagent new-msgid
				$ldpagent send-withdraw-msg $s_fec $s_lspid
			}
		}
	}
	
	# check if there is a tunnel lsp for $lspid
	set tunnelid [$self get-tunnelid-for-lspid $lspid]
	if { $tunnelid > -1 } {
                set oiface [$self get-outgoing-iface $tunnelid]
                set ldpagent [$self get-ldp-agent $oiface]
                $ldpagent new-msgid
                $ldpagent send-release-msg $fec $lspid $tunnelid
                $self out-label-clear $fec $lspid 
                return
	}
	
	$self out-label-clear $fec $lspid 
	
	set ldpagent [$self get-ldp-agent $nexthop]
	$ldpagent new-msgid
	$ldpagent send-release-msg $fec $lspid
}

# Debugging

RtModule/MPLS instproc trace-mpls {} {
        [$self set classifier_] set trace_mpls_ 1
}

RtModule/MPLS instproc pft-dump {} {
        # dump the records within Partial Forwarding Table(PFT)
        set nodeid [[$self node] id]
        [$self set classifier_] PFTdump $nodeid
}

RtModule/MPLS instproc erb-dump {} {
        # dump the records within Explicit Route information Base(ERB)
        set nodeid [[$self node] id]
        [$self set classifier_] ERBdump $nodeid
}

RtModule/MPLS instproc lib-dump {} {
        # dump the records within Label Information Base(LIB)
        set nodeid [[$self node] id]
        [$self set classifier_] LIBdump $nodeid
}

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
