#
#  Time-stamp: <2000-09-11 10:21:47 haoboy>
# 
#  Copyright (c) 1997 by the University of Southern California
#  All rights reserved.
# 
#  Permission to use, copy, modify, and distribute this software and its
#  documentation in source and binary forms for non-commercial purposes
#  and without fee is hereby granted, provided that the above copyright
#  notice appear in all copies and that both the copyright notice and
#  this permission notice appear in supporting documentation. and that
#  any documentation, advertising materials, and other materials related
#  to such distribution and use acknowledge that the software was
#  developed by the University of Southern California, Information
#  Sciences Institute.  The name of the University may not be used to
#  endorse or promote products derived from this software without
#  specific prior written permission.
# 
#  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
#  the suitability of this software for any purpose.  THIS SOFTWARE IS
#  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
#  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# 
#  Other copyrights might apply to parts of this software and are so
#  noted when applicable.
# 
#  Original source contributed by Gaeil Ahn. See below.
#

###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################

#############################################################
#                                                           #
#     File: File for Simulator class                        #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), Jan. 2000      #
#                                                           #
#############################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

Simulator instproc mpls-node args {
        $self instvar linked_mplsnodes_
	
	$self node-config -MPLS ON
	set n [$self node]
	$self node-config -MPLS OFF
	
	lappend linked_mplsnodes_ $n
	return $n
}

Simulator instproc configure-ldp-on-all-mpls-nodes {} {
        $self instvar linked_mplsnodes_
        # create ldp agents on all mpls nodes
        set len [llength $linked_mplsnodes_]
        for {set i 0} {$i < $len} {incr i} {	
		for {set k [expr $i+1]} {$k < $len} {incr k} {
			set me   [lindex $linked_mplsnodes_ $i]
			set peer [lindex $linked_mplsnodes_ $k]
			eval $self LDP-peer $me $peer
		}
        }
}

Simulator instproc LDP-peer {src dst} {
	# Establish LDP-peering between node $src and $dst. The names
	# src and dst do NOT indicate a single-direction relationship.
        if { ![$src is-neighbor $dst] } {
		return
	}
	set ldpsrc [[$src get-module "MPLS"] make-ldp]
	set ldpdst [[$dst get-module "MPLS"] make-ldp]
	$ldpsrc set-peer [$dst id]
	$ldpdst set-peer [$src id]
        $self connect $ldpsrc $ldpdst
}

Simulator instproc enable-control-driven {} {
	$self instvar linked_mplsnodes_
	# let all mpls-nodes operate as control-driven mode
	set mplslen [llength $linked_mplsnodes_]
	for {set k 0} {$k < $mplslen} {incr k} {
		# select a mpls-node
		set me [lindex $linked_mplsnodes_ $k]
		set me_module [eval $me get-module "MPLS"]
		$me_module enable-control-driven
	}
}

Simulator instproc enable-data-driven {} {
	$self instvar linked_mplsnodes_
	# let all mpls-nodes operate as data-driven mode
	set mplslen [llength $linked_mplsnodes_]
	for {set k 0} {$k < $mplslen} {incr k} {
		# select a mpls-node
		set me [lindex $linked_mplsnodes_ $k]
		set me_module [eval $me get-module "MPLS"]
		$me_module enable-data-driven
	}
}

Simulator instproc enable-on-demand {} {
	$self instvar linked_mplsnodes_
	# let all mpls-nodes operate as data-driven mode
	set mplslen [llength $linked_mplsnodes_]
	for {set k 0} {$k < $mplslen} {incr k} {
		# select a mpls-node
		set me [lindex $linked_mplsnodes_ $k]
		set me_module [eval $me get-module "MPLS"]
		$me_module enable-on-demand
	}
}

Simulator instproc enable-ordered-control {} {
	$self instvar linked_mplsnodes_
	# let all mpls-nodes operate as data-driven mode
	set mplslen [llength $linked_mplsnodes_]
	for {set k 0} {$k < $mplslen} {incr k} {
		# select a mpls-node
		set me [lindex $linked_mplsnodes_ $k]
		set me_module [eval $me get-module "MPLS"]
		$me_module enable-ordered-control
	}
}

Simulator instproc ldp-notification-color {color} {
	$self color 991 $color
}

Simulator instproc ldp-request-color {color} {
	$self color 992 $color
}

Simulator instproc ldp-mapping-color {color} {
	$self color 993 $color
}

Simulator instproc ldp-withdraw-color {color} {
	$self color 994 $color
}

Simulator instproc ldp-release-color {color} {
	$self color 995 $color
}

proc msg-append {msg val delimeter} {    
	if { [string length $msg] > 0 } {
		append msg $delimeter
	}
	append msg $val
	return $msg
}

proc msg-index {msg index delimeter} {
	set msg [split $msg $delimeter]
	set val [lindex $msg $index]
	return $val
}

proc is-exist-func {funcname} {
	set ret [lsearch [info proc] $funcname]
	return $ret
}

proc parse-bw { bspec } {
        if { [scan $bspec "%f%s" b unit] == 1 } {
                set unit b
        }
        switch $unit {
		b   { return $b }
		bps { return $b }
		k   { return [expr $b*1000] }
		K   { return [expr $b*1000] }
		kb  { return [expr $b*1000] }
		Kb  { return [expr $b*1000] }
		m   { return [expr $b*1000000] }
		M   { return [expr $b*1000000] }
		mb  { return [expr $b*1000000] }
		Mb  { return [expr $b*1000000] }
		g   { return [expr $b*1000000000] }
		G   { return [expr $b*1000000000] }
		gb  { return [expr $b*1000000000] }
		Gb  { return [expr $b*1000000000] }
		B   { return [expr $b*8] }
		kB  { return [expr $b*1000*8] }
		KB  { return [expr $b*1000*8] }
		mB  { return [expr $b*1000000*8] }
		MB  { return [expr $b*1000000*8] }
		gB  { return [expr $b*1000000000*8] }
		GB  { return [expr $b*1000000000*8] }
		default { 
			puts "error: parse-bw: unknown unit `$unit'" 
			exit 1
                }
        }
}

proc parse-delay { dspec } {
        if { [scan $dspec "%f%s" b unit] == 1 } {
                set unit s
        }
        switch $unit {
		s  { return $b }
		ms { return [expr $b*0.001] }
		ns { return [expr $b*0.000001] }
		default { 
			puts "error: parse-delay : unknown unit `$unit'" 
			exit 1
                }
        }
}

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
