###########################################################################
# Copyright (c) 2000 by Gaeil Ahn                                	  #
# Everyone is permitted to copy and distribute this software.		  #
# Please send mail to fog1@ce.cnu.ac.kr when you modify or distribute     #
# this sources.								  #
###########################################################################

#############################################################
#                                                           #
#     File: File for Reroute mechanisms                     #
#     Author: Gaeil Ahn (fog1@ce.cnu.ac.kr), June 2000      #
#                                                           #
#############################################################

#############################################################
#     RtModule/MPLS Class                                   #
#############################################################

###########################################################################
#         Adapted to ns-2.1b8 by Christian.Glomb@siemens.com              #
#         If you have problems, found bugs or know improvements,          #
#         please e-mail me (cc to Gaeil Ahn and Haobo Yu).                #
#         Thanks, Christian.                                              #
###########################################################################

Simulator instproc set-failure-type {type} {
     $self instvar linked_mplsnodes_ 

     switch $type {
            "link"       {  set type 0  }
            "node"       {  set type 1  }
            default      {  
                            puts "failure type error : $option"
                            exit
            }
    }

    set mplslen [llength $linked_mplsnodes_]
    for {set k 0} {$k < $mplslen} {incr k} {
        
          set mplsnode [lindex $linked_mplsnodes_ $k]
          
          $mplsnode set failure_type_ $type
    }
}

RtModule/MPLS instproc enable-reroute {option} {
        $self instvar classifier_ 
	
        $classifier_ set enable_reroute_ 1
        
        switch $option {
            drop                 {  $classifier_ set reroute_option_ 0  }
            L3                   {  $classifier_ set reroute_option_ 1  }
            simple-dynamic       {  $classifier_ set reroute_option_ 2  }
            notify-prenegotiated {  $classifier_ set reroute_option_ 3  }
            shortest-dynamic     {  $classifier_ set reroute_option_ 4  }
            robust-crlsp         {  $classifier_ set reroute_option_ 5  }
            default              {  
                                    puts "reroute option error : $option"
                                    exit
            }
        }
}

RtModule/MPLS instproc get-reroute-option {} {
        $self instvar classifier_ 
	
        if { [$classifier_ set enable_reroute_] != 1 } {
		return -1
        }
	
        set option [$classifier_ set reroute_option_]
        switch $option {

           0 { return "drop"                 }
           1 { return "L3"                   }
           2 { return "simple-dynamic"       }
           3 { return "notify-prenegotiated" }
           4 { return "shortest-dynamic"     }
           5 { return "robust-crlsp"         }
        }
}

RtModule/MPLS instproc enable-reroute-egress-lsr {} {
        $self instvar classifier_
        $classifier_ set reroute_egress_lsr_ 1
}

RtModule/MPLS instproc is-reroute-egress-lsr {} {
        $self instvar classifier_
        return [$classifier_ set reroute_egress_lsr_]
}

RtModule/MPLS instproc set-protection-flow {stime interval fec phb} {
	$self instvar classifier_ 
	
        if { [$classifier_ set enable_reroute_] != 1 } {
		puts "error: specify a reroute mechanism"
		exit
        }
	
        set ns [Simulator instance]
        $ns at $stime "$self check-failure $interval up $fec $phb"
}

RtModule/MPLS instproc set-protection-lsp {stime interval args} {
	$self instvar classifier_
        
        if { [$classifier_ set enable_reroute_] != 1 } {
		puts "warning: specify a reroute mechanism"
        }
	
        set ns [Simulator instance]
        set stime    [parse-delay $stime]
        set interval [parse-delay $interval]
        
        set nargs    [llength $args]
        if { $nargs == 0} {
		puts "error: specify lspid"
		exit
        }
        for {set i 0} {$i < $nargs} {incr i} {
		set lspidset [lindex $args $i]
		
		set lspidset [split $lspidset _ ]
		set len [llength $lspidset]
		
		for {set k 0} {$k < $len} {incr k} {
			set lspid [lindex $lspidset $k]
			$ns at $stime "$self check-failure $interval up $lspid"            
		}
        }
}

RtModule/MPLS instproc get-lspid-for-fecphb {fec phb} {
	$self instvar classifier_
        return [$classifier_ get-lspid-for-fecphb $fec $phb]
}

RtModule/MPLS instproc get-failnext-for-lspid {lspid} {
	$self instvar classifier_
        return [$classifier_ get-failnext-for-lspid $lspid]
}

RtModule/MPLS instproc get-faildst-for-lspid {lspid} {
	$self instvar classifier_	 
        return [$classifier_ get-faildst-for-lspid $lspid]
}

RtModule/MPLS instproc set-lib-error-for-fecphb {fec phb state} {
	$self instvar classifier_	 
        return [$classifier_ set-lib-error $fec $phb -1 $state]
}

RtModule/MPLS instproc set-lib-error-for-lspid {lspid state} {
	$self instvar classifier_
        return [$classifier_ set-lib-error -1 -1 $lspid $state]
}

RtModule/MPLS instproc schedule-events {time state fec phb lspid} {
        set ns  [Simulator instance]
        set now [$ns now]
	
        if { $lspid > -1 } {
		$ns at [expr $now + $time] "$self check-failure $time $state $lspid"
        } else {
		$ns at [expr $now + $time] "$self check-failure $time $state $fec $phb"
        }
}

RtModule/MPLS instproc lsp-setup-callback {time lspid args} {
	set ns  [Simulator instance]
	set now [$ns now]
	
	if { [$self get-outgoing-label $lspid] > -1 } {
		set nargs    [llength $args]
		for {set i 0} {$i < $nargs} {incr i} {
			append action [lindex $args $i]
			append action " "
		}
		$ns at [expr $now + $time]         $action
	} else {
		$ns at [expr $now + $time] "$self lsp-setup-callback $time $lspid $args"
	}
}

RtModule/MPLS instproc check-failure args {
        $self instvar classifier_ 
	
        set nargs    [llength $args]        
        set time  [lindex $args 0]
        set state [lindex $args 1]
                
        if { $nargs == 3 } {
		set lspid  [lindex $args 2]
		set olabel [$self get-outgoing-label $lspid]
        } elseif { $nargs == 4 } {
		set fec    [lindex $args 2]
		set phb    [lindex $args 3]
		set olabel [$self get-outgoing-label $fec $phb]
        } else {
		puts "error: wrong argument number in check-failure $args"
		exit
        }
	
        if { $olabel < 0 } {
		if { $nargs == 3 } {
			$self schedule-events $time $state -1 -1 $lspid
		} else {
			$self schedule-events $time $state $fec $phb -1
		}
		return
        }
	
        if { $nargs == 3 } {
		set oiface   [$self get-outgoing-iface $lspid]
		set fec      [$self get-fec-for-lspid  $lspid]
        } else {
		set oiface   [$self get-outgoing-iface   $fec $phb]
		set lspid    [$self get-lspid-for-fecphb $fec $phb]
        }             
	
        if { [$self get-link-status $oiface] == "down" } {
		
		if { $state == "down" } {
			if { $nargs == 3 } {
				$self schedule-events $time $state -1 -1 $lspid
			} else {
				$self schedule-events $time $state $fec $phb -1
			}
			return
		}
		set state "down"
		$classifier_ set failure_detection_ 1
		
		set option [$self get-reroute-option]
		switch $option {
			
                        drop {
                                if { $lspid > -1 } {
                                        $self set-lib-error-for-lspid $lspid 1
				} else {
					$self set-lib-error-for-fecphb $fec $phb 1
				}
			}
			simple-dynamic {
				if { $lspid > -1 } {
                                        $self set-lib-error-for-lspid $lspid 1
                                        
                                        set er [$self simple-dynamic-alg $lspid]
                                        if {$er == -1} {
						set state "up"
						if { $nargs == 3 } {
							$self schedule-events $time $state -1 -1 $lspid
						} else {
							$self schedule-events $time $state $fec $phb -1
						}
                                        }
				} else {
                                        $self set-lib-error-for-fecphb $fec $phb 1
                                        
                                        set lspid 9999
                                        $self setup-erlsp $fec $fec $lspid
                                        
                                        set time 0.001
                                        set action "$self reroute-fec-binding $fec $phb $lspid"
                                        $self lsp-setup-callback $time $lspid $action
				}
			}
			notify-prenegotiated {
				
				if { $nargs == 3 } {
					set iiface   [$self get-incoming-iface     $lspid]
				} else {
					set iiface   [$self get-incoming-iface   $fec $phb]
				}             
				
				if { $iiface > -1 } {
					set ldpagent [$self get-ldp-agent $iiface]
					$ldpagent new-msgid
					$ldpagent send-notification-msg "BSNodeError" $fec $lspid
				}
			}
			shortest-dynamic {   
				$self set-lib-error-for-lspid $lspid 1
				
				if {$lspid < 0} {
					puts "error: in check-failure"
					exit
				}
				
				set er [$self shortest-dynamic-alg $lspid]
				
				if {$er == -1} {
					set state "up"
					if { $nargs == 3 } {
						$self schedule-events $time $state -1 -1 $lspid
					} else {
						$self schedule-events $time $state $fec $phb -1
					}
					return
				}
			}
		}
        } else {
		
		if { $state == "up" } {
			if { $nargs == 3 } {
				$self schedule-events $time $state -1 -1 $lspid
			} else {
				$self schedule-events $time $state $fec $phb -1
			}
			return
		}
		set state "up"
		$classifier_ set failure_detection_ 0
		
		set option [$self get-reroute-option]
		switch $option {
			
                        drop {
                                if { $lspid > -1 } {
                                        $self set-lib-error-for-lspid $lspid 1
                                } else {
                                        $self set-lib-error-for-fecphb $fec $phb 1
                                }
                        }
			simple-dynamic {
				if { $lspid > -1 } {
                                        $self set-lib-error-for-lspid $lspid -1
                                        set lspid [expr $lspid +1]
				} else {
                                        $self set-lib-error-for-fecphb $fec $phb -1
                                        set lspid 9999
				}
				set oiface   [$self get-outgoing-iface     $lspid]
				if {$oiface > -1} {
					$self send-crldp-release-msg $lspid                                   
				}
			}
			
			notify-prenegotiated {   
				if { $nargs == 3 } {
					set iiface   [$self get-incoming-iface     $lspid]
				} else {
					set iiface   [$self get-incoming-iface   $fec $phb]
				}             
				
				if { $iiface > -1} {
                                        set ldpagent [$self get-ldp-agent $iiface]
                                        $ldpagent new-msgid
                                        $ldpagent send-notification-msg "NodeRepair" $fec $lspid
				}
			}
			
			shortest-dynamic {   
				$self set-lib-error-for-lspid $lspid -1
				
				set oiface   [$self get-outgoing-iface [expr $lspid +1]]
				if {$oiface > -1} {
					$self send-crldp-release-msg [expr $lspid +1]
				}
			}
		}
        }
	
        if { $nargs == 3 } {
		$self schedule-events $time $state -1 -1 $lspid
        } else {
		$self schedule-events $time $state $fec $phb -1
        }
}

RtModule/MPLS instproc simple-dynamic-alg { lspid } {        
	set ns [Simulator instance]
	set oiface   [$self get-outgoing-iface $lspid]
	set fec      [$self get-fec-for-lspid  $lspid]
	
	set qosid [$self get-flowid-for-lspid $lspid]
	set failnext [$self get-failnext-for-lspid $lspid]
	
	if {$failnext != -1} {
		set fnodes [split $failnext "_"]
		set fnodes [linsert $fnodes 0 $oiface]
	} else {
		set fnodes $oiface
	}
	
	set er [$ns get-shortest-path [[$self node] id] $fec $fnodes "" ]
	
	while { $er == -1 } {
		if { [llength $fnodes] == 0} {
			puts "error: simple-dynamic"
			return -1
		}
		set dst [lindex $fnodes end]
		set fnodes [lreplace $fnodes end end]
		set er [$ns get-shortest-path [[$self node] id] $dst $fnodes "" ]
	}
	
	set strlspid "L$lspid"
	set er [msg-append $er $strlspid "_"]
	#puts "=======> trying LSP setup: er=$er" 
	
	if { $qosid != -1 } {
		set nhopid [$self get-outgoing-iface $lspid]
		set linkptr [$self get-link-ptr $nhopid]
		set TRate [$linkptr get-info-in-cbq-pool $qosid bw]
		set BSize [expr [$linkptr get-info-in-cbq-pool $qosid qlim] -1]
		set PSize 1
		set SPrio [$linkptr get-info-in-cbq-pool $qosid prio]
		set HPrio $SPrio
		
		$self setup-crlsp $fec $er [expr $lspid +1] $TRate $BSize $PSize $SPrio $HPrio
	} else {
		$self setup-erlsp $fec $er [expr $lspid +1]
	}
	return $er
}

RtModule/MPLS instproc shortest-dynamic-alg { lspid } {        
	set oiface   [$self get-outgoing-iface $lspid]
	set fec      [$self get-fec-for-lspid  $lspid]
	set failnext [$self get-failnext-for-lspid $lspid]
	if {$failnext == -1} {
		set failnext $fec
	} else {
		set failnext [split $failnext "_"]
	}
	
	set ns [Simulator instance]
	set srcnode [[$self node] id]
	
	set er -1
	set cost    1000       
	
	if { [$self set failure_type_] == 0 } {
		set dstnode $oiface
		set flinks $srcnode
		lappend flinks $dstnode
		set er [$ns get-shortest-path $srcnode $dstnode "" $flinks]
		if {$er != -1} {
			set cost [llength [split $er "_"]]
		}
	}
	
	set fnodes  $oiface
	set len [llength $failnext]
	
	for {set i 0} {$i < $len} {incr i} {
		set dstnode [lindex $failnext $i]
		set newer [$ns get-shortest-path $srcnode $dstnode $fnodes ""]
		
		if { $newer != -1 } {
			set newcost [llength [split $newer "_"]]
			
			if { $newcost < $cost } {
				set er    $newer
				set cost  $newcost
			}
		}
	}
	
	if {$er == -1} {
		return $er
	}
	
	set strlspid "L$lspid"
	set er [msg-append $er $strlspid "_"]
	#puts "=======> tring LSP setup: er=$er" 
	
	set qosid [$self get-flowid-for-lspid $lspid]
	if { $qosid != -1 } {
		set nhopid [$self get-outgoing-iface $lspid]
		set linkptr [$self get-link-ptr $nhopid]
		set TRate [$linkptr get-info-in-cbq-pool $qosid bw]
		set BSize [expr [$linkptr get-info-in-cbq-pool $qosid qlim] -1]
		set PSize 1
		set SPrio [$linkptr get-info-in-cbq-pool $qosid prio]
		set HPrio $SPrio
		
		$self setup-crlsp $fec $er [expr $lspid +1] $TRate $BSize $PSize $SPrio $HPrio
	} else {
		$self setup-erlsp $fec $er [expr $lspid +1]
	}
        
	return $er
}

#############################################################
#     Simulator Class                                       #
#############################################################
Simulator instproc get-next-lsr {lsr fec nolsrid} {
	set len [llength [$lsr set neighbor_]]
	
	for {set i 0} {$i < $len} {incr i} {
		set neighbor [lindex [$lsr set neighbor_] $i]
		if { [$neighbor id] == $nolsrid } {
			continue
		}
		
		set nexthop [$neighbor get-nexthop $fec]
		if { $nexthop == [$lsr id] || $nexthop == $nolsrid } {
			continue
		}
		
		return $neighbor
	}
	
	return -1
}

Simulator instproc enable-reroute args {
	$self instvar linked_mplsnodes_ 
	# option : "drop" or "new"
	#     case drop :  If a link fails, Packets on the path are dropped.
	#     case new  :  If a link fails, try to make a new path.
	
	set nargs [llength $args]
	
	if { [llength $args] == 1 } {
		set option [lindex $args 0]
	} else {
		set option "drop"
	}
	
	set mplslen [llength $linked_mplsnodes_]
	
	for {set k 0} {$k < $mplslen} {incr k} {
		set mplsnode [lindex $linked_mplsnodes_ $k]
		set mplsmodule [$mplsnode get-module "MPLS"]
		$mplsmodule enable-reroute $option
	}
}

Simulator instproc set-protection-flow {stime interval fec phb} {
	$self instvar linked_mplsnodes_ 
	
	set mplslen [llength $linked_mplsnodes_]
	
	for {set k 0} {$k < $mplslen} {incr k} {        
		set mplsnode [lindex $linked_mplsnodes_ $k]
		set mplsmodule [$mplsnode get-module "MPLS"]
		$mplsmodule set-protection-flow $stime $interval $fec $phb
	}
}

Simulator instproc set-protection-lsp {stime interval args} {
	$self instvar linked_mplsnodes_ 
	
	set nargs    [llength $args]
	if { $nargs == 0} {
		puts "error: specify lspid"
		exit
	}
	for {set i 0} {$i < $nargs} {incr i} {
		if {$i == 0} {
			set lspidset [lindex $args $i]
		} else {
			set lspidset [msg-append $lspidset [lindex $args $i] "_"]
		}
	}
	
	set mplslen [llength $linked_mplsnodes_]
	for {set k 0} {$k < $mplslen} {incr k} {
		set mplsnode [lindex $linked_mplsnodes_ $k]
		set mplsmodule [$mplsnode get-module "MPLS"]
		$mplsmodule set-protection-lsp $stime $interval $lspidset
	}
}

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
